//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "mrbcoach.h"
#include "worldmodel.h"
#include "audiomsg.h"
#include "mrbtypes.h"
#include "minmax.h"

#include <sstream>
#include <string>

#include <stdlib.h>
#include <time.h>

using namespace std;


//#define USEAUDIOTRACE
//#define USEANALYZETRACE
//#define USEACTTRACE
//#define USETEAMTRACE

//////////////////////////////////////////////////////////////////////
//
// some local constants to determine coach behaviour
//
//////////////////////////////////////////////////////////////////////

// no of cycles in a standard situation before a second audio msg
const int AUMSG_BAN_TIME= 40;

#ifndef USEAUDIOTRACE
#define AUDIOTRACE(output)
#else
#define AUDIOTRACE(output) { ostrstream ost;\
 ost << "COACH AUDIO: " << output << (char)0;\
    char * str = ost.str(); \
    Comm::sendMCmd(MLog(str)); \
    delete str; }
#warning OnlineCoach debug trace AUDIOTRACE is activated!
#endif

#ifndef USEANALYZETRACE
#define ANALYZETRACE(output)
#else
#define ANALYZETRACE(output) { ostrstream ost;\
 ost << "COACH analyze: " << output << (char)0;\
    char * str = ost.str(); \
    Comm::sendMCmd(MLog(str)); \
    delete str; }
#warning OnlineCoach debug trace ANALYZETRACE is activated!
#endif

#ifndef USEACTTRACE
#define ACTTRACE(output)
#else
#define ACTTRACE(output) { ostrstream ost;\
 ost << "COACH act: " << output << (char)0;\
    char * str = ost.str(); \
    Comm::sendMCmd(MLog(str)); \
    delete str; }
#warning OnlineCoach debug trace ACTTRACE is activated!
#endif

#ifndef USETEAMTRACE
#define TEAMTRACE(output)
#else
#define TEAMTRACE(output) { ostrstream ost;\
 ost << "COACH act: " << output << (char)0;\
    char * str = ost.str(); \
    Comm::sendMCmd(MLog(str)); \
    delete str; }
#warning OnlineCoach debug trace TEAMTRACE is activated!
#endif


MrbCoach::MrbCoach( const char *team, const char *host, const int port,
		    const char *configfile, const char* traceCfg,
		    const bool FUNSSELmode ) :
  Onlinecoach( team, host, port, configfile, traceCfg, FUNSSELmode ),
  sayMessageNow( false ), lastPlayMode( PLAY_MODE_UNKNOWN ), 
  numMsgsLeft( SAY_COACH_CNT_MAX ), lastNewMessageTime( 0 ),
  lastAumsgTime( -3 - AUMSG_BAN_TIME )
{
  tactical     = new TacticalModule(world,NULL);

  goalShotEval = new GoalShotEval();
  passEval     = new PassEval(goalShotEval);
  positionEval = new PositionEval(goalShotEval, passEval);

  otherTeam = new Team(world);

  initSubs();
  
}

MrbCoach::~MrbCoach()
{
  delete tactical;
  delete goalShotEval;
  delete passEval;
  delete positionEval;
  delete otherTeam;

  delete [] numPlType;

}


//////////////////////////////////////////////////////////////////////
//
// act() - acting in standard situations
//
//////////////////////////////////////////////////////////////////////

void MrbCoach::act()
{
  
  ACTTRACE("Coach is acting.");
  
  // give additional messages to coach every 6000 cycles...
  if( world->time - lastNewMessageTime >= HALF_TIME_IN_CYCLES * 2 )
    {
      lastNewMessageTime = world->time;
      numMsgsLeft += SAY_COACH_CNT_MAX;
      ACTTRACE( "Added messages for additional 2 half times. Has now left: " << numMsgsLeft );
    }

  if( sayMessageNow )
    {
      ACTTRACE( "Sending audio message in a standard situation ..." );
      sayMessageNow = false;

      if( numMsgsLeft == 0 )
	cerr << "Coach: Cannot scream: Lost voice earlier!" << endl;
      else
	{
	  int player = calcPlayerForKickUs();

	  AudioMsg coachMsg( 0, world->time, true );

	  coachMsg << true << *world;                // send worldmodel
	  coachMsg << tactical->otherPlayerType();   // send opponent player types from tactical module
	  coachMsg << Int3B( player );       // send player who should do the kick
	  
	  int def1=3;
	  int def2=9;
	  otherTeam->GetDefensePlayer(def1,def2);
/*	  cerr << "Opponent's defenders: opps: opp1: " << def1 
	       << " opp2: " << def2 << endl; */
	  coachMsg  << Int3B( def1 * 12 + def2 );

	  if( !coachMsg.getState() )
	    cerr << "Coach : Fatal error: buffer overflow:" << coachMsg << endl;
	  else
	    {
	      lastAumsgTime = world->time;
	      say( coachMsg );
	      AUDIOTRACE( "Sending msg: " << coachMsg );
	      numMsgsLeft--;
	      AUDIOTRACE( "Now coach has " << numMsgsLeft << " msgs left" );
	    }
	}
    }

  if (Config::getValueAsInt("use_heterogenous_players") == 1)
      if (!initDone){
	  bool receivedTypes = true;
	  for(int i=1;i<7;i++) receivedTypes=receivedTypes && (world->playerType[i].state!=-1);
	  if (receivedTypes){
	      chooseHeterogenousPlayers();
	      initDone = true;
	  }
      }

#ifdef USEACTTRACE
  printStats();
#endif
}

void MrbCoach::chooseHeterogenousPlayers(){

    readAgentConf();

    initSubs();

    for (int i=12-tactical->ourSystem.numOffenders; i<12; i++)
	choosePlayerType(i);
    for (int i=2; i<12-tactical->ourSystem.numOffenders; i++)
	choosePlayerType(i);
    
}

void MrbCoach::choosePlayerType(const int player_nr){
    
    if ( (world->playMode == PLAY_MODE_BEFORE_KICK_OFF)
	 || (numSubs>0)){
	int type = 0;
	double utility = calcUtility(player_nr,0);
	double max_utility = utility;
	for (int i=1; i < numPlTypes; i++)
	    if (numPlType[i]>0){
		utility = calcUtility(player_nr,i);
		if (utility < 0){
		    cerr << "WARNING: utility < 0. Check agent.conf" << endl;
		}
		if (utility>max_utility){
		    max_utility = utility;
		    type = i;
		}
	    } 
	if (type != 0)
	    doSub(player_nr,type);
    }
}

double MrbCoach::calcUtility(const int player_nr,const int player_type){

  return
    desiredType[player_nr].decay_inertia_weight 
    * (1 - fabs(PLAYER_DECAY + desiredType[player_nr].decay_inertia_delta - world->playerType[player_type].playerDecay)/0.2)
    + desiredType[player_nr].kickmargin_rand_weight 
    * (1 - fabs(KICKABLE_MARGIN + desiredType[player_nr].kickmargin_rand_delta - world->playerType[player_type].kickableMargin)/0.2)
    + desiredType[player_nr].dpr_stamina_inc_max_weight
	* (1 - fabs(DASH_POWER_RATE + desiredType[player_nr].dpr_stamina_inc_max_delta - world->playerType[player_type].dashPowerRate)/0.002)
    + desiredType[player_nr].extra_stamina_effort_weight 
    * (1 - fabs(desiredType[player_nr].extra_stamina_effort_delta - world->playerType[player_type].extraStamina)/100.0);
}

void MrbCoach::readAgentConf(){
  char valueString[100];
  for (int i=2; i<12; i++){
    sprintf(valueString,"player%d_decay_inertia_delta",i);
    desiredType[i].decay_inertia_delta = Config::getValueAsDouble(valueString);
    sprintf(valueString,"player%d_decay_inertia_weight",i);
    desiredType[i].decay_inertia_weight = Config::getValueAsDouble(valueString);
    
    sprintf(valueString,"player%d_kickmargin_rand_delta",i);
    desiredType[i].kickmargin_rand_delta = Config::getValueAsDouble(valueString);
    sprintf(valueString,"player%d_kickmargin_rand_weight",i);
    desiredType[i].kickmargin_rand_weight = Config::getValueAsDouble(valueString);
    
    sprintf(valueString,"player%d_dpr_stamina_inc_max_delta",i);
    desiredType[i].dpr_stamina_inc_max_delta = Config::getValueAsDouble(valueString);
    sprintf(valueString,"player%d_dpr_stamina_inc_max_weight",i);
    desiredType[i].dpr_stamina_inc_max_weight = Config::getValueAsDouble(valueString);

    sprintf(valueString,"player%d_extra_stamina_effort_delta",i);
    desiredType[i].extra_stamina_effort_delta = Config::getValueAsDouble(valueString);
    sprintf(valueString,"player%d_extra_stamina_effort_weight",i);
    desiredType[i].extra_stamina_effort_weight = Config::getValueAsDouble(valueString);

  }


}
//////////////////////////////////////////////////////////////////////
//
// analyze() - here the playmode is used to determine whether the 
//             coach should say something now or not.
//
//////////////////////////////////////////////////////////////////////

void MrbCoach::analyze()
{ 
  tactical->feedbackWorld();
  positionEval->CalMinMaxX(world);
  positionEval->CalPlayerPositions(world);

  bool playModeToAct = false;

  if ( world->playMode==PLAY_MODE_PLAY_ON )
    {
      otherTeam->Update();
#ifdef USETEAMTRACE
      Comm::sendMCmd(POSITION(otherTeam->GetTeamCOG()));
      int i;
      for (i=2; i<12; i++)
	{
	  //	   Comm::sendMCmd(BOX(otherTeam->GetPos(i)+otherTeam->GetVar(i),
	  //			     otherTeam->GetPos(i)-otherTeam->GetVar(i)));
	  //Comm::sendMCmd(CIRCLE(otherTeam->GetPos(i),3));
	   Comm::sendMCmd(BOX(otherTeam->GetaPos(i)+otherTeam->GetaVar(i),
	  			     otherTeam->GetaPos(i)-otherTeam->GetaVar(i)));
	  Comm::sendMCmd(CIRCLE(otherTeam->GetaPos(i),3));
	};
      //     TEAMTRACE("Fitness: " << otherTeam->CalcFitness() <<" aFitness: "
      //		<< otherTeam->CalcaFitness());
#endif
    };

  switch( world->playMode )
    {
    case PLAY_MODE_KICK_OFF_US:
    case PLAY_MODE_KICK_OFF_THEM:
    case PLAY_MODE_GOAL_KICK_US:
    case PLAY_MODE_GOAL_KICK_THEM:
    case PLAY_MODE_FREE_KICK_US:
    case PLAY_MODE_FREE_KICK_THEM:
    case PLAY_MODE_KICK_IN_US:
    case PLAY_MODE_KICK_IN_THEM:
    case PLAY_MODE_CORNER_KICK_US:
    case PLAY_MODE_CORNER_KICK_THEM:
      playModeToAct= true;
      break;
    default:
      // "normal" playmodes which cannot be handled by online coach
    
      // reset the counter for the last Audiomsg:
      lastAumsgTime = -2 - AUMSG_BAN_TIME;
    }

  // did the playmode change?
  if( world->playMode != lastPlayMode )
    {
      ANALYZETRACE( "Noted Change of playmode ..." );
      lastPlayMode = world->playMode;
      if( playModeToAct )
	{
	  sayMessageNow = true;
	  ANALYZETRACE( "Set flag: scream!" );
	}
      else
	{
	  ANALYZETRACE( "Not right mode to act ..." );
	}
    }
  else   // playmode is still the same
    {
      // say something if enough time has passed since last audio message
      if( playModeToAct && world->time - lastAumsgTime > AUMSG_BAN_TIME )
	sayMessageNow = true;
    }
}


int
MrbCoach::
calcPlayerForKickUs() const
{
  int plNo = UNKNOWN;

  double minDist = 10000.0;

  for( int i = 1; i <= 11; i++ )
    {
      if( !tactical->ourTeamMidfield( i ) )
	continue;

      double plDist = norm( world->ourTeam[i].pos - world->ball.pos );
      if( plDist < minDist )
	{
	  minDist = plDist;
	  plNo = i;
	}
    }

  return plNo;
}



// which step should be used for printing statistics?
#define STATALWAYS false
// if not always then only here:
#define STATSTEP 1

void MrbCoach::doSub(const int player, const int type)
{
  // change player i to type pt
  plType[player]= type;
  if (type != 0)
    numPlType[type]--;
  
  sendChangePlayer(player, type);

#ifdef USEACTTRACE  
  { ostrstream ost;
  ost << "3 " << player << " set to type :" << type << (char)0;
  char * str = ost.str(); 
  Comm::sendMCmd(MLog(str));
  cout << str << endl;
  delete str; }


  // inform about change of player type
  { ostrstream ost;
  ost << "HetCoach: " << player << " set to type :" << type << (char)0;
  char * str = ost.str(); 
  Comm::sendMCmd(MMarker(Vector2d(player, type), ost.str())); 
  delete str; }
#endif
}

void MrbCoach::initSubs()
{
  // number of subs left in Game (after KICK_OFF !)
  // set to 0 to avoid subs in online game
  numSubs= playerParams.subs_max;
  //  static unsigned int numSubs(0);
  
  // information about how many players of one type are left
  numPlTypes= playerParams.player_types;
  numPlType= new int[numPlTypes];
  for (int i=0; i<numPlTypes; i++)
    // [C] : Max number of players of one type on the field
    numPlType[i]= 3;
  
  // initial subs not done yet!
  initDone= false;
}

void MrbCoach::printStats()
{
  if (!STATALWAYS && world->time != STATSTEP)
    return;
  
  { ostringstream ost;
  ost << "0 Coach hetero player status:";
  Comm::sendMCmd(MLog(ost.str()));}
  
  for (int i=0; i<numPlTypes; i++) {
    ostringstream ost;
    ost << "1 " << (world->playerType[i]);
    Comm::sendMCmd(MLog(ost.str()));
  }
  
  for (int i=2; i<12; i++) {
    ostringstream ost;
    ost << "2 Player " << i << " has type " << plType[i];
    Comm::sendMCmd(MLog(ost.str()));
  }
}
