//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// 	$Id: AdvSkills.h,v 1.8 2003/06/26 11:18:11 aarn Exp $
//      $Manager: aarn (Axel Arnold)$
//
// ////////////////////////////////////////////////////////////////////

#ifndef SKILLS_H
#define SKILLS_H

/** all basic skills of the player.
    All skills calculate and execute one action leading to a target given
    by the agent. They return a state from which the agent can
    decide wether another action is needed or not.
*/
class AdvSkills
{
public:
  /** @name variable constants.
      these are values that normally should not change permanently during
      a game, but that may need to be adjusted to the specific settings of
      a match or of the agent using them. They may especially depend of
      the player type (server 7.0) - and so change after the player is
      substituted! "readjustToPlayer()" will take care of that.
  */
  //@{
  /// assumed number precision. Default: 1e-5
  double precision;
  /** velocities below this are considered zero.
      Default: 0.1 */
  double minSpeed;
  /// time until that intercept times should be calculated
  long minInterceptTime;
  /// angle threshold for goToPosition. Lower values mean higher correction
  double goToThreshStretch;
  /** turn(Heads) below this are marked FINISHED.
      Default: 15 */
  double turnMinAngle;
  /// improvement threshold to choose direct kick over indirect
  double directThreshold;
  /// position error to actually lay back
  double laybackErr;
  /// the position error induced by rounding
  double data_error;
  /** the distance where the ball is handled optimally.
      Set to my.size + BALL_SIZE + 0.5 * my.kickableMargin by
      readjustToPlayer().
  */
  double safeCircle;
  /// the velocity which the ball may get maximally so that it can be handled
  double stopableV;
  /** the tolerated body direction angle error in dribble.
      Default: 10 */
  double dribbleBodyErr;
  /** the tolerated ball direction angle error in dribble.
      Default: 30 */
  double dribbleHoldErr;
  /** if dribble wants to do a dash with less than dribbleMinDash*MAXPOWER, we rather
      place.
      Default: PLAYER_RAND */
  double dribbleMinDash;
  /** if ball will be slower than this after we placed it, we rather kick it than
      place it to keep it roll.
      Default: BALL_RAND */
  double dribbleMinBallV;
  /// maximal number of steps simulated by control ball.
  long   controlMaxSteps;
  //@}

  /***************************** the constructor ****************************/
  /** the constructor.
    @param comm used for sending the calculated commands.
    @param w is player's worldmodel
    */
  AdvSkills(Comm *comm, Worldmodel *w);

  /******************* re-adjustment of variable constants ******************/

  void readjustToPlayer(const WPlayer &my);


  /********************************** the skills ****************************/

  /// kicking commands
  //@{
  /// compatability - translate modes into time lengths
  enum KickMode { KICKAWAY = 1, FIXED = 2, SUPERKICK = 5 };

  /**
    \begin{itemize}
    \item[KRANGE]     ball out of kick range
    \item[KFINISHING] ball will have wanted velocity after this kick
    \item[KTOOSHORT]  ball will be slower than wanted
    \item[KPLACING]   ball is placed to allow a better kick
    \item[KPLACEFORSKICK] ball is layed back
    \item[KSTOPPING]  requested velocity is near 0, so stopping ball
    \end{itemize}
   */
  enum KickState { KRANGE, KFINISHING, KTOOSHORT, KPLACING,
		   KPLACEFORSKICK, KSTOPPING };
  ///
  struct KickCommand {
    /// calculated ball velocity
    Vector2d  v;
    ///
    KickState state;
  };

  struct ExtKickCommand: public KickCommand {
    /// how many steps it will likely cost to complete the kick
    long time;
    ///
    long max_time;
    ///
    Vector2d target;
    /// if >= 0, the target velocity, indicates position kick
    double target_v;
  };

  /** try to give the ball a target velocity. Collisions are avoided by kicking
      the ball around the player if necessary. It is possible to let kick try
      to place the ball behind the player for a really hard kick (superkick).
    
      It is possible, that you have to call kick more than once. So call
      {\em rekick} to continue the last kick until it says RANGE,
      so that as long as possible kicks are sent.

      @param target_v      the velocity the ball should get
      @param max_time      maximal time until kick must be finished
      */
  ExtKickCommand kick(const Vector2d &target_v, long max_time = 2);

  /** kick to an absolute target position.
      @param target_pos    the absolute coordinates where the ball should
                           be kicked.
      @param target_v      the velocity the ball should have when reaching
                           the target position
      */
  ExtKickCommand kick(double target_v, const Vector2d &target_pos, long max_time = 2);

  /** does the last kick command again.
      @param kcm           the KickCommand from kick to continue
      */
  ExtKickCommand rekick(const ExtKickCommand &kcm);

  /** try to place ball on a safe circle around the player. The balls velocity
      will be slow enough to be stopped within one turn, if possible.
      @param target        the wanted position of the ball relative to the
                           player in absolute coordinates.
      */
  KickCommand placeBall(const Vector2d &target);

  /** try to stop ball.
      */
  KickCommand stopBall();

  /// execute part for kick commands
  void exec(const KickCommand &);
  //@}

  /// intercepting skills
  //@{
  /**
    \begin{itemize}
    \item[TFINISHED]       at target position/ball
    \item[TONWAY]          player is on way, but not in target position
    \item[TGOINGBACKWARDS] player shows in opposite direction
    \end{itemize}
   */
  enum TurtleState { TFINISHED, TONWAY, TGOINGBACKWARDS };
  ///
  struct TurtleCommand {
    /// estimated time until finished
    long        time;
    /// calculated player velocity
    Vector2d    v;
    /// players new body dir
    Angle       bodyDir;
    /// the angle to turn to reach bodyDir (affected by inertia moment)
    double      turnMoment;
    ///
    TurtleState state;
    // command storage
    enum { DASH, TURN } command;
  };

  /** try to get the ball under control, that is, intercept it.
      @param stamina       the stamina to use at most this step
      @param meet          the position we will meet the ball
      */
  TurtleCommand controlBall(double stamina, double error, WBall &meet);
  ///
  TurtleCommand controlBall(double stamina);

  /** go to a position. This is intended for long range intercepts and absolute
      positions, so error is assumed to be relatively large ( > player size).
      TIME INFORMATION OF TURTLECOMMAND IS NOT SET!!!

      @param target        absolute target position on field
      @param error         radius around the target in which the player
                           should end up
      @param stamina       the stamina to use at most this step
      @param dashBackwardsIfNeeded  keep the difference between bodyDir and
                           prefAngle < 90 by eventually dashing backwards
      @param prefAngle     the angle to keep if dashBackwardsIfNeeded is set
      */
  TurtleCommand goToPosition(const Vector2d &target,
			     double error = 5.0,
			     double stamina = STAMINA_INC_MAX,
			     bool   dashBackwardsIfNeeded = false,
			     Angle  prefAngle = 0);
  /** run in a direction. This is basicly for interceptions of opponents, who 
      are already behind our defense. 
      TIME INFORMATION OF TURTLECOMMAND IS NOT SET!!!

      @param rightdir      right limit for direction
      @param leftdir       left limit for direction
      @param prefAngle     angle to turn to, if a turn is needed      
      @param stamina       the stamina to use at most this step
      */
  TurtleCommand runInDir(double rightdir,
			 double leftdir,
			 double prefAngle = -9999,
			 double stamina = MAXPOWER);

  /** go to a position.
      @param target        absolute target position on field
      @param stamina       the stamina to use for the complete intercept additional to
                           the stamina gained during the intercept.
      @param error         radius around the target in which the player
                           should end up
   */
  TurtleCommand goToPositionTimed(const Vector2d &target,
				  double stamina, double error);

  /** try to intercept given WObject. Lowlevel stuff behind goToPositionTimed and
      controlBall.
      @param my            the player
      @param target        the target
      @param stamina       the stamina to use at most this step
      @param meet          where we meet the target
      */
  TurtleCommand interceptTimed(const WPlayer &my,
			       const WMovedObject &target,
			       double stamina, double error,
			       WMovedObject &meet);
  
  ///
  void exec(const TurtleCommand &);
  //@}

  /// dribbling
  //@{
  /**
    \begin{itemize}
    \item[DOK]           simply dribbling
    \item[DGOINGFORBALL] ball lost, actually intercepting it
    \end{itemize}
   */
  enum DribbleState { DOK, DGOINGFORBALL };
  ///
  struct DribbleCommand {
    /// calculated player and ball velocity
    Vector2d     v_p, v_b;
    ///
    Angle        bodyDir;
    ///
    DribbleState state;
    /// command storage
    enum { DASH, TURN, KICK } command;
  };

  /** dribble into an absolute direction. The agent can decide in which
      direction to the player the ball should be kept (in absolute direction).

      @param dribbledir    the direction in which to dribble
      @param hold_dir      the direction we want to hold the ball from us 
      */
  DribbleCommand dribble(Angle dribbledir,
			 Angle hold_dir, double stamina = STAMINA_INC_MAX);
  ///
  void exec(const DribbleCommand &);
  //@}

  /// goalie's catch skill
  //@{
  /**
    \begin{itemize}
    \item[CBAN]       catch still banned
    \item[CRANGE]     ball to far away or outside penalty area
    \item[CNOTGOALIE] player is not goalie and may not catch
    \item[COK]        done
    \end{itemize}
    */
  enum CatchState { CBAN, CRANGE, CNOTGOALIE, COK };

  ///
  struct CatchCommand {
    ///
    Angle catchDir;
    ///
    CatchState state;
  };

  /** catch the ball in absoulte dir "dir", if allowed. This is checked via worldmodel.
      */
  CatchCommand goalieCatch( Angle dir );
  ///
  void exec(const CatchCommand &);
  //@}

  /// turn
  //@{
  /**
    \begin{itemize}
    \item[ROK]       done
    \item[RANGLE]    angle is too large, truncated
    \item[RFINISHED] angle to turn is small
    \end{itemize}
   */
  enum TurnState{ ROK, RANGLE, RFINISHED };
  ///
  struct TurnCommand {
    /// the optimal body dir actually achievable
    Angle bodyDir;
    /// the angle to turn to reach bodyDir (affected by inertia moment)
    double turnMoment;
    ///
    TurnState state;
  };
  /** turn into absolute direction dir.
      */
  TurnCommand turn(Angle dir);

  /** turn body towards object obj.
      */
  TurnCommand turnTo(const WObject &obj);
  ///
  void exec(const TurnCommand &);
  //@}

  /// turn neck
  //@{
  /**
    \begin{itemize}
    \item[THOK]       done
    \item[THANGLE]    angle is too large, truncated
    \item[THFINISHED] angle to turn small
    \item[THBAN]      no command set. Call illegal
    \end{itemize}
  */
  enum TurnHeadState{ THOK, THANGLE, THFINISHED, THBAN };
  ///
  struct TurnHeadCommand {
    /// the optimal face dir actually achievable
    Angle  faceDir;
    double turnHeadMoment;
    /// desired view width
    view_width vwidth;
    ///
    TurnHeadState state;
  };

  /** turns head into specified absolute direction. The view width
      is set to NORMAL, but this may be altered for exec.

      @param dir          the absolute dir we want to see
      */
  TurnHeadCommand turnHead(Angle dir);

  /** turns head to some object. The view width
      is set to NORMAL, but this may be altered for exec.

      @param dir          the absolute dir we want to see
      */
  TurnHeadCommand turnHead(const WObject &obj);

  ///
  void exec(const TurnHeadCommand &);
  /** sets the turn "angle" that is assumed for the turnHead calculations.
      Do not use if you are not absolutely sure what that does. RTFS.
  */
  void   enterExpTurnHeadTurn(double turnParam = 0);
  ///
  void   leaveExpTurnHeadTurn();
  ///
  double getTurnHeadTurn() const { return turnParam; }

  /// block use of turnHead. Allowed after next exec call.
  void   blockTurnHead() { blockTH = true; }
  ///
  bool   turnHeadBlocked() const { return blockTH; }
  //@}

  /// point to an absolute position
  //@{
  /**
    \begin{itemize}
    \item[PTBAN]       pointo still banned
    \item[PTOK]        done
    \end{itemize}
    */
  enum PointToState { PTBAN, PTOK };

  ///
  struct PointToCommand {
    ///
    Angle dir;
    ///
    double dist;
    ///
    bool off;
    ///
    PointToState state;
  };

  /** point to a position, if allowed. This is checked via worldmodel.
      @param pos the position to point to in absolute coordinates
      */
  PointToCommand pointTo(const Vector2d &pos);
  /** stop pointing to a position, if allowed. This is checked via
      worldmodel.
      */
  PointToCommand pointToOff();
  ///
  void exec(const PointToCommand &);
  //@}

  /// tackling commands
  //@{
  /**
    \begin{itemize}
    \item[TOK]        ball will have wanted velocity after this tackle
    \item[TKICKABLE]  ball is in kickrange
    \item[TRANGE]     ball out of tackle range or failProb > 50%
    \item[TTOOSHORT]  ball will be slower than wanted
    \item[TTOOFAST]   ball is to fast to be kicked to given dist
                      (assuming dist = target_v)
    \end{itemize}
   */
  enum TackleState { TOK, TKICKABLE, TRANGE, TTOOSHORT, TTOOFAST };
  ///
  struct TackleCommand {
    /// calculated kickpower
    double power;
    /// probability ($\in [0,1]$) that tackle will fail
    double failProb;
    /// The direction the ball is tackled to
    Vector2d dir;
    ///
    TackleState state;
  };

  /** calculate the probability of a tackle command
      @return the probability ($\in [0,1]$) that the tackle command will succeed
   */

  double tackleProb();

  /** tackle ball away as fast as possible
      */

  TackleCommand tackle();

  /** tackle ball away
      @param dist          how far the ball should be pushed.
      @param target_v      the velocity the ball should have when reaching
                           the target position
      */
  TackleCommand tackle(const double &dist, const double &target_v);
  ///
  void exec(const TackleCommand &);
  //@}

  /** @name low-level commands.
      These are all commands that are directly available from comm.C.
   */
  //@{
  ///
  void noAction();
  ///
  void attentionTo(int pNum, bool ourTeam = true);
  ///
  void attentionToOff(void);

  ///
  void basicDash (double pow);
  ///
  void basicKick (double pow, double dir);
  ///
  void basicTurn (double pow);
  ///
  void basicCatch(double dir);
  ///
  void basicMove(double x, double y);
  ///
  void basicPointTo(double dist, double dir);
  ///
  void basicPointToOff(void);
  ///
  void basicTackle(double power);

  ///
  void sendTurnNeck(double dir);
  ///
  void sendChangeView(view_width width, view_quality quality);
  ///
  void sendSenseBody(void);
  //@}

private:

  Comm *comm;
  Worldmodel *world;

  // turn head stuff (taking into account turns)
  double turnParam;
  bool blockTH;
  // buffer
  double btParam;
  bool   bblockTH;
  bool   expmode;

  // for pointto
  double turnHeadParam;

  // calculate current max and min accelerations for dashing respecting
  // reachability
  double reachRadius(double stamina, int numDashes, double decay);
  double goToThresh(double error, double tdist);

  // vel constraints
  void calcAcc(double stamina, double &minacc, double &maxacc);
  // internal placeBall with more options
  KickCommand iplaceBall(const Vector2d &target, double collRange,
			 int cond_mask, const int *drop_order);
};

extern AdvSkills *skills;

#endif

