//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// cwz-build-command: make all

/* basic algorithm: we turn in the direction we want to dribble,
   then dash. Besides we look what might go wrong (ball lost!)
   and care for that.
*/
AdvSkills::DribbleCommand
AdvSkills::dribble(Angle ddir,
		   Angle hdir, double stamina)
{
  DribbleCommand com;
  com.state = DOK;

  const WPlayer &my(world->my);
  const WBall   &ball(world->ball);

  Angle ddirmor = ddir - my.bodyDir;

  // simulated positions next step
  Vector2d pns = my.pos + my.v, bns = ball.pos + ball.v;

  // minimal and maximal allowed ball positions
  double max_random_error = M_SQRT2*(norm(world->my.v)*PLAYER_RAND +
				     norm(world->ball.v)*BALL_RAND) + data_error;
  double collRange = BALL_SIZE + world->my.size + max_random_error;
  double maxRange = my.kickRange() - max_random_error;

  // possible dashes
  double maxdash = my.dashPowerRate * max(min(stamina, MAXPOWER), 0.0);
  Vector2d dashVec = radAng2Vector2d(maxdash, my.bodyDir);
  // planned action
  bool doturn = false, contBall = false, fearCollision = false;
  // planned dash
  double dashlen = 0;

  // tgtvec is the wanted ball relative position
  Vector2d tgtvec = radAng2Vector2d(min(safeCircle, maxRange), hdir);

  if (fabs(ddirmor) > dribbleBodyErr) {
    // calc for turn
    LEVELTRC(SKILLS_LEVEL, "dribble: planning to turn");
    doturn = true;
  }
  else {
    // calc for dash: as fast as possible, but NO collision
    double llow, lhigh;
    switch (intersectLineCircle(-dashVec, pns - bns, 1.05*collRange, llow, lhigh)) {
    case None:
      // ok dashing will not hit the ball, so do what we can
      dashlen = 1;
      break;
    case Valid:
      // take the longer dash, if possible
      if (lhigh <= 1)
	dashlen = lhigh;
      else if (llow <= 1)
	dashlen = llow;
      else
	dashlen = 1;
      // dashing to slow makes no sense
      if (dashlen < dribbleMinDash)
	contBall = true;
      break;
    default:
      cerr << "AdvSkills: Unexpected result of intersectLineCircle!!!!!" << endl;
      break;
    }
    dashVec *= dashlen;
    pns += dashVec;
    LEVELTRC(SKILLS_LEVEL, "dribble: planning to dash");
    LEVELDRW(SKILLS_LEVEL, MDLine(my.pos, pns));
    LEVELDRW(SKILLS_LEVEL, MDLabel(pns + dashVec, "planned dash"));
  }

  // now check wether we have to care for the ball rather
  if (!contBall) {
    // check for collisions
    if (norm(my.pos - bns) < collRange) {
      // collision by ball movement first => place it away
      LEVELTRC(SKILLS_LEVEL, "dribble: collision by ball move");
      contBall = fearCollision = true;
    }
    else if (norm(pns - bns) < collRange) {
      // collision by second movement => place ball away
      LEVELTRC(SKILLS_LEVEL, "dribble: collision by player move");
      contBall = fearCollision = true;
    }
    else if (norm(pns - bns) > maxRange) {
      // ball will be lost
      LEVELTRC(SKILLS_LEVEL, "dribble: ball will be lost");
      contBall = true;
    }
    else if (sqrnorm(ball.v) > sqrnorm(my.v) &&
	     norm(bns - (pns + tgtvec)) >= my.decay*my.vmax) {
      /* most sophisticated reason: if the ball will be too fast
	 if we have to place it back next step, rather kick now
	 otherwise we will have to kick next, then stop, which
	 results in a stopped and therefore too slow ball. */
      LEVELTRC(SKILLS_LEVEL, "dribble: ball will be too fast");
      contBall = true;
    }
  }

  if (contBall) {
    // place the ball to relative position tgtvec NEXT time step
    KickCommand kcm = placeBall(my.v + tgtvec);
    if (kcm.state == KRANGE) {
      com.state = DGOINGFORBALL;
      // oops, somehow we lost the ball, go get it
      TurtleCommand tcm = controlBall(stamina);
      switch (tcm.command) {
      case TurtleCommand::TURN:
	com.command = DribbleCommand::TURN;
	break;
      case TurtleCommand::DASH:
	com.command = DribbleCommand::DASH;
      default:;
      }
      com.v_p     = tcm.v;
      com.v_b     = world->ball.v;
      com.bodyDir = tcm.bodyDir;
      return com;
    }
    
    /* if we will more or less nail the ball to its position,
       we rather try to kick it a little bit into the dribble
       direction and catch it again. Nailing means that the ball
       has more or less its correct relative position and that it
       has a bad speed that will make us loose it. So we drive it
       into the dribble direction.
       Even more complicated: If we plan to turn, this is probably
       not a good idea (CM).
    */
    if (norm(com.v_b) < dribbleMinBallV) {
      if (!doturn) {
	LEVELTRC(SKILLS_LEVEL, "dribble: modified kick to keep up speed");
	// sustained velocity with current stamina and one cycle correction
	// possible
	kcm = kick(radAng2Vector2d(my.decay*maxdash/(1-my.decay), ddir), 1);
      }
      else {
	// if we turn, make the ball actually STOP to enable turning next round.
	// but not if we fear a collision
	if (!fearCollision)
	  kcm = placeBall(ball.pos - my.pos);
      }
    }
    com.command = DribbleCommand::KICK;
    com.v_p     = world->my.v;
    com.v_b     = kcm.v;
    com.bodyDir = world->my.bodyDir;
    return com;
  }

  if (doturn) {
    com.command = DribbleCommand::TURN;
    com.v_p     = world->my.v;
    com.v_b     = world->ball.v;
    TurnCommand tmptcm = turn(ddir);
    com.bodyDir = tmptcm.bodyDir;
    return com;
  }

  LEVELDRW(SKILLS_LEVEL, MDPosition(pns));
  LEVELDRW(SKILLS_LEVEL, MDCircle(pns, maxRange));
  com.command = DribbleCommand::DASH;
  com.v_p     = world->my.v + dashVec;
  com.v_b     = world->ball.v;
  com.bodyDir = world->my.bodyDir;
  return com;
}

void AdvSkills::exec(const DribbleCommand &com) {
  switch (com.command) {
  case DribbleCommand::KICK: {
    LEVELTRC(SKILLS_LEVEL, "dribbleanalyze: kick");
    Vector2d kick_v = com.v_b - world->ball.v;
    double res_power = norm(kick_v) / world->my.kickPowerRate(world->ball);
    Angle res_angle = kick_v.getAngle() - world->my.bodyDir;
    // renormation necessary for precision reasons
    if (res_power > MAXPOWER)
      res_power = MAXPOWER;
    basicKick(res_power, res_angle);
    break;
  }
  case DribbleCommand::TURN:
    LEVELTRC(SKILLS_LEVEL, "dribbleanalyze: turn");
    basicTurn(com.bodyDir - world->my.bodyDir);
    break;
  case DribbleCommand::DASH:
    LEVELTRC(SKILLS_LEVEL, "dribbleanalyze: dash");
    basicDash(norm(com.v_p - world->my.v)/world->my.dashPowerRate);
    break;
  default:
    cerr << world->time << "AdvSkills::exec(DribbleCommand): called without valid command" << endl;
    noAction();
  }  
}
