//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/************************************ kick *********************************/

// cwz-build-command: make all

/*
  KICK policy:
  KICKAWAY: always one kick
  FIXED: placeBall until ball kickable; then always kick.
  SUPERKICK: check if direct kick is ok, else lay back first. Then
             continue in FIXED mode.
 */

// flags for satisfied conditions in iplaceBall
enum { C_MA = 1, C_ST = 2, C_SA = 4, C_C1 = 8, C_C2 = 16 };

AdvSkills::ExtKickCommand
AdvSkills::kick(double tgv, const Vector2d &tpos, long max_time)
{
  if (tgv < 0)
    tgv = 0;

  ExtKickCommand com;
  com.target   = tpos;
  com.target_v = tgv;
  com.max_time = max_time;

  return rekick(com);
}

AdvSkills::ExtKickCommand
AdvSkills::kick(const Vector2d &tpos, long max_time)
{
  ExtKickCommand com;
  com.max_time = max_time;
  com.target   = tpos;
  double n = norm(com.target);
  if (n > BALL_SPEED_MAX)
    com.target.stretch(BALL_SPEED_MAX/n);
  com.target_v = -1;
  return rekick(com);
}

AdvSkills::ExtKickCommand
AdvSkills::rekick(const ExtKickCommand &tgtcom)
{
  // for iplace without stopableSpeed check
  const int drop_order[] = { C_SA, C_C2, C_C1, C_MA};
  const int init_mask = C_SA|C_C2|C_C1|C_MA;

  ExtKickCommand com;
  
  // setup continuation copies
  com.target   = tgtcom.target;
  com.target_v = tgtcom.target_v;
  
  // for position kicking, calculate velocity vector
  Vector2d target = tgtcom.target;
  if (tgtcom.target_v >= 0) {
    target -= world->ball.pos;
    double nrm = norm(target);
    if (nrm) {
      double tgv = tgtcom.target_v + nrm*(1-BALL_DECAY);
      if (tgv > BALL_SPEED_MAX)
	tgv = BALL_SPEED_MAX;
      target.stretch(tgv/nrm);
    }
  }

  const WBall   &ball(world->ball);
  const WPlayer &my(world->my);

  if (!my.ballKickable(ball)) {
    LEVELDRW(SKILLS_LEVEL, MLog("kick: RANGE"));
    com.state = KRANGE;
    return com;
  }

  LEVELDRW(SKILLS_LEVEL, MDLine(ball.pos, ball.pos + target));

  // parameters
  double maxacc = my.kickPowerRate(ball)*MAXPOWER;
  double collRange = BALL_SIZE + my.size + M_SQRT2*(norm(my.v)*PLAYER_RAND + norm(ball.v)*BALL_RAND);

  // check one kick
  Vector2d k1;
  Vector2d diff = target - ball.v;
  double diffnrm = norm(diff);
  if (diffnrm < maxacc) {
    k1 = target;
  }
  else {
    // calculate one kick not correct, but optimal
    double l1, l2;
    switch(intersectLineCircle(target, ball.v, maxacc, l1, l2)) {
    case Valid:
      if (l1 > 1)
	k1 = l1*target;
      else
	k1 = l2*target;
      break;
    default:
      k1 = ball.v + (maxacc/diffnrm)*diff;
      break;
    }
  }
  // modify to not collide if necessary
  if (sqrnorm(ball.pos + k1 - my.pos) < sqr(collRange)) {
    KickCommand pcm = iplaceBall(ball.pos - my.pos + k1,
				 collRange, init_mask, drop_order);
    k1 = pcm.v;
  }
  double k1err = sqrnorm(k1 - target);

  // com keeps the one step less solution
  com.state = KTOOSHORT;
  com.v     = k1;
  com.time  = 1;
  com.max_time = 1;
  if ((tgtcom.max_time <= 1) ||
      (sqrnorm(com.v - target) < sqr(directThreshold))) {
    LEVELDRW(SKILLS_LEVEL, MDLine(ball.pos, ball.pos + com.v));
    LEVELDRW(SKILLS_LEVEL, MDLabel(ball.pos + com.v, "1 kick solution, perfect"));
    return com;
  }

  LEVELDRW(SKILLS_LEVEL, MDLine(ball.pos, ball.pos + com.v));
  LEVELDRW(SKILLS_LEVEL, MDLabel(ball.pos + com.v, "1 kick solution, not perfect"));

  // two kicks. get ball as near as possible and in front
  Angle frontdir = (target + BALL_DECAY*(ball.pos - my.pos)).getAngle();
  // now ignore stopableSpeed (we want to kick, right?)
  KickCommand pcm2 =
    iplaceBall(radAng2Vector2d(safeCircle, frontdir),
	       collRange, init_mask, drop_order);

  // check what kicking there results in
  WPlayer smy = my;
  WBall sball = ball;
  smy.simulate();
  sball.v = pcm2.v;
  sball.simulate();
  double maxacc2 = smy.kickPowerRate(sball)*MAXPOWER;

  diff    = target - sball.v;
  diffnrm = norm(diff);
  if (diffnrm < maxacc2) {
    // after two kicks, we are finished
    com.time  = 2;
    com.max_time = 2;
    com.state = KPLACING;
    com.v = pcm2.v;
    LEVELDRW(SKILLS_LEVEL, MDLabel(ball.pos + com.v, "2 kick solution, perfect"));
    LEVELDRW(SKILLS_LEVEL, MDLine(ball.pos, ball.pos + com.v));
    return com;
  }

  Vector2d k2 = sball.v + (maxacc2/diffnrm)*diff;
  double k2err = sqrnorm(k2 - target);

  LEVELDRW(SKILLS_LEVEL, MDLabel(ball.pos + k2, "2 kick solution, not perfect"));
  LEVELDRW(SKILLS_LEVEL, MDLine(ball.pos, ball.pos + com.v));
  LEVELDRW(SKILLS_LEVEL, MDLine(ball.pos + com.v, ball.pos + k2));

  if (k1err > k2err + directThreshold) {
    // 2 kicks better, update com
    com.state = KPLACING;
    com.time  = 2;
    com.max_time = 2;
    com.v     = pcm2.v;
  }
  
  if (tgtcom.max_time == 2)
    return com;
  
  // here we know that 2 steps are not enough; use three steps
  // place ball in opposite frontdir if too fast for now or not there
  Vector2d laybackpos = radAng2Vector2d(collRange + 0.5*BALL_SIZE, PI + target.getAngle());
  // here be careful about stopableSpeed since somewhat laying back
  KickCommand pcm3 = placeBall(laybackpos);
  // only place if kick is really doing something
  if (sqrnorm(pcm3.v - ball.v) > minSpeed) {
    com.time  = tgtcom.max_time - 1; // take a little time for placing
    com.max_time  = tgtcom.max_time - 1;
    com.state = KPLACING;
    com.v     = pcm3.v;

    LEVELDRW(SKILLS_LEVEL, MDLabel(Vector2d(0, 0), "trying place behind"));
    return com;
  }

#ifdef MRBDEBUG
  if (GETCONF()->draw) {
    switch (com.time) {
    case 1:
      LEVELDRW(SKILLS_LEVEL, MDLabel(Vector2d(0, 0), "using 1 kick solution"));
      break;
    case 2:
      LEVELDRW(SKILLS_LEVEL, MDLabel(Vector2d(0, 0), "using 2 kick solution"));
      break;
    default:
      LEVELDRW(SKILLS_LEVEL, MDLabel(Vector2d(0, 0), "using obscure kick solution"));
      break;
    }
  }
#endif

  // return old 2/1 step command
  return com;
}

AdvSkills::KickCommand
AdvSkills::stopBall()
{
  KickCommand com;
  // stop the ball
  if (!world->my.ballKickable(world->ball)) {
    com.state = KRANGE;
    return com;
  }

  double ns = norm(world->ball.v);
  double kickrate = world->my.kickPowerRate(world->ball);
  com.v = (ns > MAXPOWER*kickrate) ? (1 - MAXPOWER*kickrate/ns)*world->ball.v : Vector2d(0, 0);
  com.state = KSTOPPING;
  return com;
}

///////////////////////////////////////////// placeBall stuff starts here


#ifdef MRBDEBUG
static void printFlag_(const Vector2d &pos, const char *txt, int f)
{
  char buf[256];
  strcpy(buf, txt);
  strcat(buf, ": (|");
  if (f & C_MA)
    strcat(buf, "MA|");
  if (f & C_SA)
    strcat(buf, "SA|");
  if (f & C_C1)
    strcat(buf, "C1|");
  if (f & C_C2)
    strcat(buf, "C2|");
  if (f & C_ST)
    strcat(buf, "ST|");
  strcat(buf, ")");
  LEVELDRW(SKILLS_LEVEL, MDLabel(pos, buf));
}
#endif

static inline void printFlag(const Vector2d &pos, const char *txt, int f)
{
#ifdef MRBDEBUG
  if (GETCONF()->draw)
    printFlag_(pos, txt, f);
#endif
}


AdvSkills::KickCommand
AdvSkills::placeBall(const Vector2d &targ)
{
  // this is the order in which conditions are dropped if necessary
  // after dropping the last all points are valid, and tgt is always aviable
  const int drop_order[] = { C_ST, C_SA, C_C1, C_C2, C_MA};
  double collRange = BALL_SIZE + world->my.size + M_SQRT2*(norm(world->my.v)*PLAYER_RAND + norm(world->ball.v)*BALL_RAND);

  return iplaceBall(targ, collRange, C_ST|C_SA|C_C1|C_C2|C_MA, drop_order);
}
 
AdvSkills::KickCommand
AdvSkills::iplaceBall(const Vector2d &targ, double collRange, int cond_mask, const int *drop_order)
{
  KickCommand com;
  const WPlayer &my(world->my);
  const WBall &ball(world->ball);

  if (!my.ballKickable(ball)) {
    com.state = KRANGE;
    return com;
  }

  // ok, the one or other way we place the ball
  com.state = KPLACING;
  
  const Vector2d ptb = ball.pos - my.pos;
  const Vector2d ptbns = ptb + ball.v;
  const Vector2d pnstb = ptb - my.v;
  const Vector2d pnstbns = pnstb + ball.v;
  const double kickrate = my.kickPowerRate(ball);
  const double maxacc = kickrate*MAXPOWER;
  const Vector2d vt = targ - pnstb;

  LEVELDRW(SKILLS_LEVEL, MDLine(ball.pos, ball.pos + vt));
  LEVELDRW(SKILLS_LEVEL, MDCircle(ball.pos, stopableV));
  LEVELDRW(SKILLS_LEVEL, MDLabel(ball.pos + radAng2Vector2d(stopableV, -0.5*M_PI), "st"));
  LEVELDRW(SKILLS_LEVEL, MDCircle(ball.pos + ball.v, maxacc));
  LEVELDRW(SKILLS_LEVEL, MDLabel(ball.pos + ball.v + radAng2Vector2d(maxacc, -0.5*M_PI), "ma"));
  LEVELDRW(SKILLS_LEVEL, MDCircle(my.pos + my.v, safeCircle));
  LEVELDRW(SKILLS_LEVEL, MDLabel(my.pos + my.v + radAng2Vector2d(safeCircle, -0.5*M_PI), "sa"));
  LEVELDRW(SKILLS_LEVEL, MDCircle(my.pos, collRange));
  LEVELDRW(SKILLS_LEVEL, MDLabel(my.pos + radAng2Vector2d(collRange, M_PI), "c1"));
  LEVELDRW(SKILLS_LEVEL, MDCircle(my.pos + my.v, collRange));
  LEVELDRW(SKILLS_LEVEL, MDLabel(my.pos + my.v + radAng2Vector2d(collRange, 0.5*M_PI), "c2"));

  // ok, now we have to find the best approximation.
  // Simply calculate all critical points and test them until one good is found
  // often needed test constants for comparison
  double sinns = sqr(collRange) - precision; // compared down
  double smaxa = sqr(maxacc) + precision;
  double ssafe = sqr(safeCircle) + precision;
  double sstop = sqr(stopableV) + precision;

  // intersecting pts of circles ma(xacc), st(opable), sa(fe) and c1/c2(collrange)
  // no c2sa since collRange < safeCircle (better is this)
  static const int mast(0), mac1(2), mac2(4), masa(6), stsa(8), stc1(10),
    stc2(12), c1sa(14), c1c2(16);
  // and the nonconstrained critical points (including free)
  static const int ma(18), st(19), c1(20), c2(21), sa(22), tgt(23), numPoints(24);
  static char *names[numPoints] = {
    "mast_l", "mast_r", "mac1_l", "mac1_r", "mac2_l", "mac2_r", "masa_l", "masa_r",
    "stsa_l", "stsa_r", "stc1_l", "stc1_r", "stc2_l", "stc2_r", "c1sa_l", "c1sa_r",
    "c1c2_l", "c1c2_l",
    "ma", "st", "c1", "c2", "sa", "tgt"
  };

  Vector2d critpoint[numPoints];
  // cond_mask gives all conditions that have to be satisfied
  int flag[numPoints];
  // later some may be dropped

  for (int i = 0; i < numPoints; i++)
    flag[i] = 0;

  // first set tgt to vt and set flags (for evtl. dropping conditions)
  // just to avoid errors from copy and paste...
  int curPnt = tgt;
  critpoint[curPnt] = vt;
  if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
    flag[curPnt] |= C_MA;
  if (sqrnorm(critpoint[curPnt]) <= sstop)
    flag[curPnt] |= C_ST;
  if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
    flag[curPnt] |= C_SA;
  if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
    flag[curPnt] |= C_C1;
  if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
    flag[curPnt] |= C_C2;

  if (flag[curPnt] == cond_mask) {
    // short cut, hopefully often used
    com.v = critpoint[curPnt];
    return com;
  }

  // calc ma/st intersecting points
  curPnt = mast;
  if (intersectCircleCircle(stopableV, ball.v, maxacc,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    flag[curPnt] = flag[curPnt + 1] = C_MA|C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb - critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;
  }

  // calc ma/c1 intersecting points
  curPnt = mac1;
  if (intersectCircleCircle(collRange, ptbns, maxacc,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_MA|C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;
  }

  // calc ma/c2 intersecting points
  curPnt = mac2;
  if (intersectCircleCircle(collRange, pnstbns, maxacc,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= pnstb;
    critpoint[curPnt + 1] -= pnstb;

    flag[curPnt] = flag[curPnt + 1] = C_MA|C_C2;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;
  }

  // calc ma/sa intersecting points
  curPnt = masa;
  if (intersectCircleCircle(safeCircle, pnstbns, maxacc,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= pnstb;
    critpoint[curPnt + 1] -= pnstb;
    
    flag[curPnt] = flag[curPnt + 1] = C_MA|C_SA;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;
    
    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;
  }

  // calc st/sa intersecting points
  curPnt = stsa;
  if (intersectCircleCircle(safeCircle, pnstb, stopableV,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= pnstb;
    critpoint[curPnt + 1] -= pnstb;

    flag[curPnt] = flag[curPnt + 1] = C_ST|C_SA;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;

    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;
  }

  // calc st/c1 intersecting points
  curPnt = stc1;
  if (intersectCircleCircle(collRange, ptb, stopableV,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_ST|C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;
 
    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;
  }

  // calc st/c2 intersecting points
  curPnt = stc2;
  if (intersectCircleCircle(collRange, pnstb, stopableV,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= pnstb;
    critpoint[curPnt + 1] -= pnstb;

    flag[curPnt] = flag[curPnt + 1] = C_ST|C_C2;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;
 
    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(ptb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C1;
    if (sqrnorm(ptb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C1;
  }

  // calc c1/sa intersecting points
  curPnt = c1sa;
  if (intersectCircleCircle(collRange, my.v, safeCircle,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_C1|C_SA;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;
  }

  // calc c1/sa intersecting points
  curPnt = c1sa;
  if (intersectCircleCircle(collRange, my.v, safeCircle,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_C1|C_SA;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) >= sinns)
      flag[curPnt    ] |= C_C2;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) >= sinns)
      flag[curPnt + 1] |= C_C2;
  }

  // calc c1/c2 intersecting points
  curPnt = c1c2;
  if (intersectCircleCircle(collRange, my.v, collRange,
			    critpoint[curPnt],
			    critpoint[curPnt + 1]) == Valid) {
    critpoint[curPnt    ] -= ptb;
    critpoint[curPnt + 1] -= ptb;

    flag[curPnt] = flag[curPnt + 1] = C_C1|C_C2;

    if (sqrnorm(critpoint[curPnt    ]) <= sstop)
      flag[curPnt    ] |= C_ST;
    if (sqrnorm(critpoint[curPnt + 1]) <= sstop)
      flag[curPnt + 1] |= C_ST;

    if (sqrnorm(critpoint[curPnt    ] - ball.v) <= smaxa)
      flag[curPnt    ] |= C_MA;
    if (sqrnorm(critpoint[curPnt + 1] - ball.v) <= smaxa)
      flag[curPnt + 1] |= C_MA;

    if (sqrnorm(pnstb + critpoint[curPnt    ]) <= ssafe)
      flag[curPnt    ] |= C_SA;
    if (sqrnorm(pnstb + critpoint[curPnt + 1]) <= ssafe)
      flag[curPnt + 1] |= C_SA;
  }

  // calc projection onto ma-circle
  curPnt = ma;
  Vector2d proj = vt - ball.v;
  double nrmprj = norm(proj);
  if (nrmprj > maxacc) {
    flag[curPnt] = C_MA;
    critpoint[curPnt] = ball.v + (maxacc/nrmprj)*proj;

    if (sqrnorm(critpoint[curPnt]) <= sstop)
      flag[curPnt] |= C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
      flag[curPnt] |= C_SA;

    if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C2;
  }

  // calc projection onto co-circle
  curPnt = c1;
  proj = ptb + vt;
  nrmprj = norm(proj);
  if ((nrmprj < collRange) && (nrmprj > precision)) {
    flag[curPnt] = C_C1;
    critpoint[curPnt] = (collRange/nrmprj)*proj - ptb;

    if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
      flag[curPnt] |= C_MA;

    if (sqrnorm(critpoint[curPnt]) <= sstop)
      flag[curPnt] |= C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
      flag[curPnt] |= C_SA;

    if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C2;
  }

  // calc projection onto co-circle
  curPnt = c2;
  proj = pnstb + vt;
  nrmprj = norm(proj);
  if ((nrmprj < collRange) && (nrmprj > precision)) {
    flag[curPnt] = C_C2;
    critpoint[curPnt] = (collRange/nrmprj)*proj - pnstb;

    if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
      flag[curPnt] |= C_MA;

    if (sqrnorm(critpoint[curPnt]) <= sstop)
      flag[curPnt] |= C_ST;

    if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
      flag[curPnt] |= C_SA;

    if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C1;
  }

  // calc projection onto sa-circle
  curPnt = sa;
  proj = pnstb + vt;
  nrmprj = norm(proj);
  if ((nrmprj < safeCircle - precision) ||
      (nrmprj > safeCircle + precision)) {
    flag[curPnt] = C_SA;
    critpoint[curPnt] = (safeCircle/nrmprj)*proj - pnstb;

    if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
      flag[curPnt] |= C_MA;

    if (sqrnorm(critpoint[curPnt]) <= sstop)
      flag[curPnt] |= C_ST;

    if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C2;
  }

  // calc projection onto st-circle
  curPnt = st;
  nrmprj = norm(vt);
  if (nrmprj > stopableV) {
    flag[curPnt] = C_ST;
    critpoint[curPnt] = (stopableV/nrmprj)*vt;

    if (sqrnorm(critpoint[curPnt] - ball.v) <= smaxa)
      flag[curPnt] |= C_MA;

    if (sqrnorm(ptb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C1;

    if (sqrnorm(pnstb + critpoint[curPnt]) >= sinns)
      flag[curPnt] |= C_C2;

    if (sqrnorm(pnstb + critpoint[curPnt]) <= ssafe)
      flag[curPnt] |= C_SA;
  }

  // and now search for a valid solution
  int drop = 0;
  while (cond_mask != 0) {
    // optimize over all points matching condition
    double min = 1e30;
    int    mindex = -1;
    for (int i = 0; i < numPoints; i++) {
      if ((flag[i] & cond_mask) == cond_mask) {
	double dist = sqrnorm(critpoint[i] - vt); 
	if (dist < min) {
	  min = dist;
	  mindex = i;
	}
      }
    }
    if (mindex >= 0) {
      // got a point
      com.v = critpoint[mindex];

      printFlag(ball.pos + critpoint[mindex], names[mindex], flag[mindex]);
      LEVELDRW(SKILLS_LEVEL, MDLine(ball.pos, ball.pos + com.v));

      return com;
    }
    // drop next condition
    cond_mask &= ~drop_order[drop++];
  }
  // never reached
  cerr << "AdvSkills::placeBall: didn't find a point to place!" << endl;
  // at least a valid kick command...
  com.v = flag[ma] ? critpoint[ma] : vt;
  return com;
}

void AdvSkills::exec(const AdvSkills::KickCommand &com) {
  if (com.state == KRANGE) {
    cerr << world->time << " AdvSkills::exec(KickCommand): ball too far away to kick!" << endl;
    noAction();
    return;
  }
  Vector2d kick_v = com.v - world->ball.v;
  double res_power = norm(kick_v) / world->my.kickPowerRate(world->ball);
  Angle res_angle = kick_v.getAngle() - world->my.bodyDir;
  // renormation necessary for precision reasons
  if (res_power > MAXPOWER)
    res_power = MAXPOWER;
  basicKick(res_power, res_angle);
}
