//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "coach.h"
#include "Config.h"

using namespace std;

#undef TRACE
#define TRACE(cmd)


Coach::Coach (const char*  team_name,             
	      playing_side side,
	      const char*  host,
	      int          port,
	      const char*  configfile,
	      bool         logging_world,
	      bool         FUNSSELmode) :
  
  world(NULL),
  comm(NULL)
{
  TRACE (cout << "Coach: initializing config reader" << endl);
  
  // try to initialize config file
  if (!Config::readFile(configfile)) {
    cerr << "PANIC: Could not read config file: " << configfile << endl;
    exit(-1);
  }

  TRACE (cout << "Coach: generating Comm" << endl);
    
  comm = new Comm(team_name, 
		  CLIENT_TRAINCOACH,
		  host, port,
		  0,
		  FUNSSELmode,
		  side);

  comm_error ce;
  if (comm->isServerAlive()) {
    TRACE (cout << "Coach: connecting" << endl);
    ce = comm->connect();
  }
  else
    ce = COMM_ERROR;
    
  switch (ce) {
  case (COMM_OK) : 
      
    TRACE (cout << "Coach: generating Worldmodel" << endl);
    
    world = new Worldmodel (comm);

    writeWorld = false;
    if (logging_world) {
      writeWorld = true;
      // open world_log file
      ostringstream tmp;
      tmp << WORLD_LOG_FILENAME << getMyTeam() << getMyNumber() 
	  << WORLD_LOG_EXTENSION;
      const char* filename = tmp.str().c_str();
      out.open(filename, ios::out);
      // Version control
      out << WORLDLOG_PROTOCOL << endl;
    }
    break;
  case (COMM_ERROR): 
    cerr << "Coach: unspecified error when initializing !" << endl; break;
  case (COMM_ERROR_SERVER_NOT_FOUND):
    cerr << "Coach: could not find server's host !"<< endl; break;
  case (COMM_ERROR_SOCKET_INIT):
    cerr << "Coach: could not create and bind socket !" << endl; break;
  case (COMM_ERROR_SERVER_INIT):
    cerr << "Coach: could not deliver checkin message to server !" 
	 << endl; break;
  case (COMM_ERROR_PARSE):
    cerr << "Coach: could not parse init response !" << endl; break;
  case (COMM_ERROR_NO_DATA):
    cerr << "Coach: did not receive response from server !" << endl; break;
  }
  
}


Coach::~Coach () {

  TRACE (cout << "Coach: entering destructor, closing down" << endl);

  delete (world);
  delete (comm);

  if (writeWorld) out.close();

  TRACE (cout << "Coach: leaving destructor" << endl);

}


void 
Coach::coach() {

  if (comm->isServerAlive()) {
    TRACE (cout << "Coach: started" << endl);

    while (comm->isServerAlive() && !comm->isGameOver() ) {
      
      TRACE (cout << "Coach: waiting for signals and handling them" << endl);
      comm->drainSocket();

      if (comm->isActionNeeded()) {
	
	TRACE (cout << "Coach: adjusting worldmodel" << endl);
	world->adjust();
      
	if ( writeWorld && comm->isTimeRunning() ) {
	  world->writeToStream(out);
	  out << endl;
	}

	TRACE (cout << "Coach: acting" << endl);
	act();
      }
	
      else {
	TRACE(cout << "Coach: no new timestep yet" << endl);
      }

    } // while server alive

    TRACE (cout << "Coach: server is gone, leaving" << endl);
    
  } else { 
    // server never was alive
    cerr << "Coach: could not connect to sserver, closing down" << endl;
  }
  
} //coach


void
Coach::changeMode(play_mode pm) {
  comm->sendChangeMode(pm);
}

void
Coach::changePlayerType(int player, int type) {
  comm->sendChangePlayer(player, type);
}

void
Coach::ear(bool earOn) {
  comm->sendEar(earOn);
}

void
Coach::moveObject (const char *object, double x, double y) {
  comm->sendMoveObject(object, x, y);
}

void
Coach::moveObject (const char *object, double x, double y, 
		   double dir_rad) {
  comm->sendMoveObject(object, x, y, dir_rad);
}

void
Coach::moveObject (const char *object, double x, double y, 
		   double dir_rad, double xvel, double yvel) {
  comm->sendMoveObject(object, x, y, dir_rad, xvel, yvel);
}

void
Coach::recover() {
  comm->sendRecover();
}
