//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef _WORLDMODEL_H_
#define _WORLDMODEL_H_

#include "sensor.h"        // sensor data types
#include "coordinates.h"   // angle class and vector class for 2-dimensional space
#include "constants.h"     // constants used by soccerserver
#include "minmax.h"        // min, max, sqr
#include "Comm.h"          // communication module
#include "audiomsg.h"      // audio messages


// constants needed by worldmodel classes

const int  UNKNOWN = -1;       // this is used for unknown player numbers and types
const int  DEFAULT_TYPE = 0;   // default player type
const long ETERNAL = 32000;    // age of objects we have never seen (->invalid objects)
                               // the value should be larger than match length
const long MAXAGE  = 20;       // objects older than this are (by default) not
                               // considered in the sorting methods
const int  VISUALSTEPS = 10;   // number of visual info directions to be stored


// utility functions concerning positions of objects on the field

// test if "pos" is in field (with security distance)
bool inField( const Vector2d &pos, double security = 0.0 );
// test if "pos" is in our penalty area (with security distance)
bool inOurPenaltyArea( const Vector2d &pos, double security = 0.0 );
// test if "pos" is in opponent penalty area (with security distance)
bool inOtherPenaltyArea( const Vector2d &pos, double security = 0.0 );

// utility functions concerning quantization
double lowerDistLimit( double nodaDist, double qstep );
double upperDistLimit( double nodaDist, double qstep );

// Reports if there could have been a player<->ball collision when simulating a cycle.

enum CollisionState
{
  NONE,        // There was no collision (according to our data) or we have new visual data
  CRITICAL,    // There MUST have been a collision, at least according to our
               // simulated data, but simulation of this collision might be wrong
               // because soccerserver uses random values in this case (very rare!)
  PROCESSED    // There WAS a collision (acc. to our data) AND it could be simulated
};


// collects data about a player substitution (player number and timestep)

struct Substitution
{
  int  number;
  long time;
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WObject
//
// any object on the field
//
///////////////////////////////////////////////////////////////////////////////////////////////

class WObject {
public:
  WObject( const Vector2d& _pos = Vector2d() ) :
    pos( _pos ), age( ETERNAL ), state( UNSAFE ) {}

  Vector2d  pos;   // absolute position on the field
  long      age;   // number of cycles that have passed since this object
                   // has been updated for the last time

  WObjectState state;   // tells which data type object was simulated from

  // calculate distance to a given object "obj"
  double    distTo( const WObject& obj ) const { return dist( pos, obj.pos ); }

  // calculate the distance of the object to the straight line between
  // source and target
  // if the projection of the object onto the line source->target
  // is not a point on this line the function returns the smaller of
  // the dist to source and the dist to target
  double    distToLine( const Vector2d& source, const Vector2d& target ) const;

  // check if object is in corridor as sketched below:
  //              ___________________________________ __         _
  //             /                                   |  \         |
  //            /                                    |   \        |
  // source -> /_ _ _ _ _ _ _ _ _ _ _ _ _ target -> _|_   |        \ d
  //           \                                     |    |        /
  //            \                                    |   /        |
  //             \___________________________________|__/        _|
  //
  // On the right there is a half circle with diameter d
  // (which is difficult to draw with ASCII characters...).
  // The angle of the cone on the left is 2 * alpha.
  // 
  // CAUTION: it is also checked if object is in a very small circle around source!
  bool      inCorridor( Vector2d source, Vector2d target, double d, Angle alpha = PI_2 ) const;

  bool      inField( double security = 0.0 ) const;
  bool      inOurPenaltyArea( double security = 0.0 ) const;
  bool      inOtherPenaltyArea( double security = 0.0 ) const;
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WMovedObject
//
// any moving object on the field
//
///////////////////////////////////////////////////////////////////////////////////////////////

class WPlayer;

class WMovedObject: public WObject {
public:
  WMovedObject( double _decay = 0.0,
		double _vmax = 0.0,
		double _size = 0.0,
		const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
		const Vector2d& _v = Vector2d( 0.0, 0.0 ) ) :
    WObject( _pos ), v( _v ), decay( _decay ), vmax( _vmax ), size( _size ), coneTime( 0 ) {}

  // accelerate object, but consider maximum velocity
  void      acc( const Vector2d& _a )
  { v += _a; if( v.getLength() > vmax ) v *= vmax / v.getLength(); }

  // simulate one moving step
  void      simulate() { pos += v; v *= decay; }

  // calculate object's absolute velocity if distChange and dirChange relative to 
  // relObject (usually player himself) are given
  void      calcVelocity( const WMovedObject& relObject, double distChange, double dirChange );

  // update object position with sensor data (== relative position to "my")
  void      setPos( const SensorObject& object, const WPlayer& my );

  Vector2d  v;          // velocity vector
  double    decay;      // um... that's how we brake (decay factor)
  double    vmax;       // maximum LENGTH of velocity vector
  double    size;       // object radius
  long      coneTime;   // number of times object position was in view cone but object unseen
  // (does not make sense without connection to a player)
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WBall
//
///////////////////////////////////////////////////////////////////////////////////////////////

class WBall : public WMovedObject {
public:
  WBall( const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
	     const Vector2d& _v = Vector2d( 0.0, 0.0 ) ) :
    WMovedObject( BALL_DECAY, BALL_SPEED_MAX, BALL_SIZE, _pos, _v ) {}

  // checks if the ball (with its current speed) will cross our goal line
  // at y-coordinates between -distance and +distance
  // (use, e.g., distance = 0.5 * GOAL_WIDTH to check if the ball will enter the goal)
  bool willCrossOurGoalLine( double distance = 0.5 * GOAL_WIDTH );
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WPlayer
//
///////////////////////////////////////////////////////////////////////////////////////////////

class WPlayer : public WMovedObject {
public:
  WPlayer( const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
	   const Vector2d& _v = Vector2d( 0.0, 0.0 ) ) :
    WMovedObject( PLAYER_DECAY, PLAYER_SPEED_MAX, PLAYER_SIZE, _pos, _v ),
    bodyAge( ETERNAL ),
    bodyDir( 0.0 ),
    faceDir( 0.0 ),
    pointDir( 0.0 ),
    number( UNKNOWN ),
    isGoalie( GOALIE_UNKNOWN ),
    tacticalNo( UNKNOWN ),
    cyclesSinceTackle( ETERNAL ),
    stamina( STAMINA_MAX ),
    effort( 1.0 ),
    recovery( 1.0 ),
    staminaIncMax( STAMINA_INC_MAX ),
    inertiaFactor( INERTIA_MOMENT ),
    dashPowerRate( DASH_POWER_RATE ),
    kickableMargin( KICKABLE_MARGIN ),
    kickRand( KICK_RAND ),
    extraStamina( 0.0 ),
    effortMin( EFFORT_MIN ),
    effortMax( EFFORT_MAX ),
    type( UNKNOWN )
  {
    if( dashPowerRate * MAXPOWER / ( 1.0 - decay ) < vmax )
      vmax = dashPowerRate * MAXPOWER / ( 1.0 - decay );
  }

  // calculate angle between player's BODY direction and "obj"
  Angle      bodyAngleTo( const WObject& obj ) const
  { return ( obj.pos - pos ).getAngle() - bodyDir; }

  // calculate angle between player's FACE direction and "obj"
  Angle      faceAngleTo( const WObject& obj ) const
  { return ( obj.pos - pos ).getAngle() - faceDir; }

  // calculate player's kick range
  double     kickRange() const { return BALL_SIZE + size + kickableMargin; }

  // check if ball is in kickable area
  bool       ballKickable( const WBall& ball ) const
  { return distTo( ball ) < kickRange() ? true : false; }

  // check if ball is in catchable area (for a catch command with catchAngle)
  bool       ballCatchable( const WBall& ball, const Angle& catchAngle ) const;

  // calculate actual kick power from angle and distance to ball
  double     kickPowerRate( const WBall& ball ) const;

  // calculate inertia moment for turn command
  double     inertiaMoment() const { return v.getLength() * inertiaFactor + 1; }

  // calculate probability (in [0,1]) that tackle command will succeed
  double     tackleProb( const WBall& ball ) const;

  long       bodyAge;          // age of player's bodyDir (and usually faceDir and velocity)

  Angle      bodyDir;          // player's body-and-running direction
  Angle      faceDir;          // player's facing direction ( == bodyDir + headAngle )
  Angle      pointDir;         // direction player is pointing to

  Angle      headAngle;        // angle of player's head relative to the body (from senseBody)

  int        number;           // player's number ( -1 == undefined, 0 == coach )
  goalie_player isGoalie;      // is player goalie?
  bool       isCoach() const   // is player not a field player but the coach?
  { return ( number == 0 ); }

  // player's tactical number in the tactical system we currently play (can be UNKNOWN!)
  // --------------|--------------------------------------------------------------------
  // role          |  tactical number
  // --------------|--------------------------------------------------------------------
  // goalie        |  1
  // defenders     |  2 .. numDefenders + 1
  // def. midfield |  numDefenders + 2 .. numDefenders + numDefMidfield + 1
  // off. midfield |  numDefenders + numDefMidfield + 2 .. 10 - numOffenders
  // midfield      |  numDefenders + 2 .. numDefenders + numMidfield + 1
  // offenders     |  11 - numOffenders .. 11
  // -----------------------------------------------------------------------------------
  int        tacticalNo;

  long       cyclesSinceTackle;   // cycles passed since player tackled for the last time
  double     stamina;             // player's stamina
  double     effort;              // player's effort (which shows how "tired" player is)
  double     recovery;            // recovery value for calculation of player's stamina

  // Parameters needed since heterogenuous players have been introduced
  double     staminaIncMax;
  double     inertiaFactor;       // CAUTION: This is called "inertia moment" by the server!
  double     dashPowerRate;
  double     kickableMargin;
  double     kickRand;
  double     extraStamina;
  double     effortMin;
  double     effortMax;

  int        type;             // player type (-1 == UNKNOWN, 0 == DEFAULT_TYPE, 1...)

  // update player with sensor data (== relative position to "my")
  void       setBySensorData( SensorPlayer player,
			      const WPlayer& my,
			      WObjectState newState,
			      play_mode playMode = PLAY_MODE_UNKNOWN );
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WPlayerType
//
///////////////////////////////////////////////////////////////////////////////////////////////

class WPlayerType {
public:
  WPlayerType() :
    state( UNKNOWN ), playerSpeedMax( PLAYER_SPEED_MAX ),
    playerDecay( PLAYER_DECAY ), playerSize( PLAYER_SIZE ),
    staminaIncMax( STAMINA_INC_MAX ), inertiaFactor( INERTIA_MOMENT ),
    dashPowerRate( DASH_POWER_RATE ), kickableMargin( KICKABLE_MARGIN ),
    kickRand( KICK_RAND ), extraStamina( 0.0 ),
    effortMin( EFFORT_MIN ), effortMax( EFFORT_MAX )
    {}

  int    state;    // -1 == UNKNOWN, 0 == DEFAULT_TYPE, player's type otherwise

  double playerSpeedMax;
  double playerDecay;
  double playerSize;
  double staminaIncMax;
  double inertiaFactor;
  double dashPowerRate;
  double kickableMargin;
  double kickRand;
  double extraStamina;
  double effortMin;
  double effortMax;
};

class std::ostream & operator << ( class std::ostream &os, const WPlayerType &pt );


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class WAction
//
// step-bound actions player can perform (turn, catch, kick, dash, tackle, move)
//
///////////////////////////////////////////////////////////////////////////////////////////////

class WAction {
public:
  WAction() 
    : command( NO_COMMAND ), moment( 0.0 ), power( 0.0 ), pos ( 0.0, 0.0 ) {}

  CycleCommand  command;    // different types of actions (see "Comm.h")

  // parameters for different commands

  double        moment;   // valid if type == TURN_COMMAND or CATCH_COMMAND or KICK_COMMAND
  double        power;    // valid if type == DASH_COMMAND or KICK_COMMAND or TACKLE_COMMAND
  Vector2d      pos;      // valid if type == MOVE_COMMAND

  void          store( const CycleAction &newAction );   // fill with sensor data
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class Worldmodel
//
///////////////////////////////////////////////////////////////////////////////////////////////

class Worldmodel {
public:
  Worldmodel( Comm * _comm );
  
  Worldmodel( const Worldmodel& wm );
  
  Worldmodel& Worldmodel::operator = ( const Worldmodel& wm );
  
  ~Worldmodel();

  // Objects on the field and related data

  WPlayer          my;               // myself

  WPlayer          ourTeam[12];      // players of my team sorted by number (0 stands for coach)
  WPlayer          otherTeam[12];    // players of opponent team sorted by number

  int              ourGoalieNo;      // number of our goalie (-1 == UNKNOWN)
  int              otherGoalieNo;    // number of opponent goalie

  int              maxPlayersOurTeam;     // for testing: limits number of players in our team (->adjustPlayers)
  int              maxPlayersOtherTeam;   // same for opponent team

  WBall            ball;
  
  WObject          ourGoal;
  WObject          otherGoal;

  CollisionState   collState;        // reports player<->ball collisions (after simulation)

  // Data related to the chronology of the game

  long             time;             // world time (time of latest senseBody information)
  long             visualAge;        // number of cycles since latest visual information

  long             cyclesSinceCatch; // no. of cycles passed since last catch command
  long             cyclesSincePointTo; // no. of cycles passed since last pointTo command (currently unused)

  playing_side     ourSide;
  playing_side     penaltySide;
  play_mode        playMode;

  int              ourScore;         // number of goals we scored
  int              otherScore;       // number of goals scored by opponent team

  WAction          action;           // player's action for current cycle

  WPlayerType*     playerType;       // the player types generated by the server
  Substitution*    ourSubs;          // the substitutions we made
  int              numOurSubs;       // ...and the number of them
  Substitution*    otherSubs;        // the substitutions the opponent made
  int              numOtherSubs;     // ...and the number of them

  long             kickCount;        // numbers of commands server has received
  long             dashCount;
  long             turnCount;
  long             sayCount;
  long             turnNeckCount;
  long             catchCount;       // the following three are new in soccerserver 7
  long             moveCount;
  long             changeViewCount;
  long             tackleCount;      // the last three are new in soccerserver 8
  long             pointToCount;
  long             attentionToCount;

  view_width       viewWidth;        // angle of view cone
  view_quality     viewQuality;      // quality of visual information

  // calculate view angle (== HALF of view cone!) from thisVwidth
  Angle            getViewAngle( view_width thisVwidth ) const;

  Angle            vDir[VISUALSTEPS];      // collects directions player received visual info from...
  Angle            vAngle[VISUALSTEPS];    // ...and the view angle when he received this information...
  long             vAge[VISUALSTEPS];      // ...and the number of cycles passed since he received this information
  long             vCounter;               // last used position in vDir and vAge arrays

  Angle            aOppDir[VISUALSTEPS];      // collects directions player received audial info from...
  long             aOppAge[VISUALSTEPS];      // ...and the number of cycles passed since he received this information
  long             aOppCounter;               // last used position in aOppDir arrays

  Angle            aOurDir[VISUALSTEPS];      // collects directions player received audial info from...
  long             aOurAge[VISUALSTEPS];      // ...and the number of cycles passed since he received this information
  long             aOurCounter;               // last used position in aOurDir arrays

  // age of a direction (no. of cycles that have passed since last visual info from that direction)
  long             dirAge( const Angle& alpha ) const;

  // returns true if the worldmodel is not too old
  // unused at the moment (and it is only a HACK!)
  int lookAroundFinished( ) const;

  // Get the number of that player of our resp. other team whose rank is "index" if the
  // players are sorted by distance to "target". Default for "index" is 1 (closest player to
  // "target"). Consider only players younger than "maxAge". Considering players we have never
  // seen before (which for example could mean that their pos is still marked as zero!) can be
  // avoided by setting maxAge < ETERNAL.
  // If no suitable player is found, UNKNOWN is returned. (!)
  // CAUTION: player himself might be returned by ourTeamClosestTo()!

  int     ourTeamClosestTo( const Vector2d& target, long maxAge = MAXAGE, int index = 1 ) const;
  int     otherTeamClosestTo( const Vector2d& target, long maxAge = MAXAGE, int index = 1 ) const;

  // count players in our resp. other team that are in a certain corridor between "source" and
  // "target" (for details of the corridor see WObject::inCorridor())
  // Consider only players with age <= maxAge.
  // CAUTION: player himself will be in the corridor and thus counted by ourTeamInCorridor()!

  int     ourTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			     const Angle& alpha = PI_2, long maxAge = MAXAGE ) const;
  int     otherTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			       const Angle& alpha = PI_2, long maxAge = MAXAGE ) const;
  
  // calculate minimum x-position that is judged to be offside
  // Only opponent players with age <= maxAge are considered
  double  ourOffsideLine( long maxAge ) const;

  // calculate the opponent's offside line (of interest for our defenders)
  double  otherOffsideLine( long maxAge ) const;

  // for soccerserver 8+ communication
  void    encodeAudio( MessageType msgType, char* msg, WBall comBall );
  // [MW] returns true if player's position was told
  bool    decodeAudio( char* msg, int sender, double senderDir );

  // write/read worldmodel non-binary to/from stream
  bool    writeToStream( std::ostream &out ) const;
  bool    readFromStream( std::istream &in );

  SensorData     adjust();   // update worldmodel with new data (if there is)
                             // returns teammate's audio message (if there is one)

  // feed audio information (from coach) into worldmodel
  bool    feedbackAudio( AudioMsg &audioMsg );

private:
  Vector2d oldBallRelpos;   // needed sometimes for guess of ball velocity
  WPlayer  oldMy;           // we need position and bodyDir from the last cycle

  bool    waitForSenseBody; // needed for senseBody synchronisation
 
  bool    adjustExact( const SensorData &data );   // fill in new exact data (from coach)

  // tries to determine view width of a see message (fills in data.see_view_width)
  bool    checkViewWidth( SensorData &data );

  // calculate my absolute position and face direction with new sensor data.
  // Face direction is calculated with a line, if only one line is seen
  // (this means we are inside the field), otherwise with flags.
  // Position is calculated with flags.
  // calcMyPos() is true if position could be calculated.
  bool    calcMyPos( const SensorData& data );

  bool    adjustPlayers( const SensorData &data );   // adjust other players to my (new) position
  bool    adjustBall( SensorData data );             // same with ball
  bool    adjustGoals( const SensorData &data );     // same with goals (only checks age)

  void    adjustTacticalNumbers( int newGoalieNo );

  Comm*   comm;         // communication module
  
  void    simCycle();   // simulate one cycle
};


class std::ostream& operator << ( class std::ostream &os, Worldmodel& w );

class AudioMsg& operator << ( class AudioMsg& aum, const Worldmodel& w );


#endif
