//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "goalshoteval.h"


//////////////////////////////////////////////////////////////////////
//
// class GoalShotEval
//
//////////////////////////////////////////////////////////////////////

GoalShotEval::GoalShotEval()
{
  // edges of opponent goal
  leftEdgeOfGoal  = Vector2d( 0.5 * FIELD_LENGTH, -0.5 * GOAL_WIDTH );
  rightEdgeOfGoal = Vector2d( 0.5 * FIELD_LENGTH, 0.5 * GOAL_WIDTH );

  // edges of our goal
  leftEdgeOfOurGoal  = Vector2d( -0.5 * FIELD_LENGTH, 0.5 * GOAL_WIDTH );
  rightEdgeOfOurGoal = Vector2d( -0.5 * FIELD_LENGTH, -0.5 * GOAL_WIDTH );

  // read constants from agent.conf

  // minimum and maximum distances to the edges of the goal
  minDistToCorner = Config::getValueAsDouble( "goal_minDistToCorner" );
  maxDistToCorner = Config::getValueAsDouble( "goal_maxDistToCorner" );

  maxAge = Config::getValueAsInt( "goal_maxAge" );

  maxKickDuration = Config::getValueAsInt( "goal_maxKickDuration" );

  score_POOR    = Config::getValueAsInt( "goal_score_POOR" );
  score_AVERAGE = Config::getValueAsInt( "goal_score_AVERAGE" );
  score_GOOD    = Config::getValueAsInt( "goal_score_GOOD" );
  score_PERFECT = Config::getValueAsInt( "goal_score_PERFECT" );

  maxSimSteps   = Config::getValueAsInt( "goal_maxSimSteps" );

  maxDist = BALL_SPEED_MAX * ( power( BALL_DECAY, maxSimSteps ) - 1.0 ) / ( BALL_DECAY - 1.0 );
};


//////////////////////////////////////////////////////////////////////
//
// class GoalShotEval::ourEval
//
//////////////////////////////////////////////////////////////////////

Grade GoalShotEval::ourEval( const Worldmodel * world, const Vector2d &playerpos ) const
{
  Vector2d target;
  return teamEval( world, playerpos, target, true );
};


//////////////////////////////////////////////////////////////////////
//
// class GoalShotEval::otherEval
//
//////////////////////////////////////////////////////////////////////

Grade GoalShotEval::otherEval( const Worldmodel * world, const Vector2d &playerpos ) const
{
  Vector2d target;
  return teamEval( world, playerpos, target, false );
}


//////////////////////////////////////////////////////////////////////
//
// class GoalShotEval::teamEval
//
//////////////////////////////////////////////////////////////////////

Grade GoalShotEval::teamEval( const Worldmodel * world, const Vector2d &playerpos,
			      Vector2d &bestTarget, bool ourTeamGoalShot ) const
{
  Grade grade;

  double playerDist[12];   // distance opponent players can dash during siumlation
  double playerV[12];      // opponent players' velocities
  bool   simPlayer[12];    // simulate this player?

  Vector2d leftTarget;
  Vector2d rightTarget;

  if( ourTeamGoalShot )
    {
      leftTarget = leftEdgeOfGoal + correctCorner( dist( playerpos, leftEdgeOfGoal ) );
      rightTarget = rightEdgeOfGoal - correctCorner( dist( playerpos, rightEdgeOfGoal ) );
   }
  else
    {
      leftTarget = leftEdgeOfOurGoal - correctCorner( dist( playerpos, leftEdgeOfOurGoal ) );
      rightTarget = rightEdgeOfOurGoal + correctCorner( dist( playerpos, rightEdgeOfOurGoal ) );
    }

  Vector2d target;     // for different goalshot targets

  WBall simBall;       // simulated ball

  long numSteps = 0;   // number of simulation steps

  long score    = score_POOR + 1;   // score for current target
  long minScore = score_POOR + 1;   // best (minimum) score
  grade.set_BAD();

  bestTarget = 0.5 * ( leftTarget + rightTarget );   // initialize with middle of goal
  
  // if player is far from goal cancel all calculations
  if( dist( playerpos, bestTarget ) > maxDist )
    return grade;

  for( int i = 1; i <= 11; i++ )
    {
      simPlayer[i] = true;
      const WPlayer & player = ourTeamGoalShot ? world->otherTeam[i] : world->ourTeam[i];
      if( player.age >= MAXAGE || dist( player.pos, bestTarget ) > maxDist )
	simPlayer[i] = false;
    }

  // test different goalshot directions
  for( int direction = 0; direction < 5; direction++ )
    {
      target = leftTarget + ( direction / 4.0 ) * ( rightTarget - leftTarget );
      simBall.pos = playerpos;
      simBall.v.setXY( 0.0, 0.0 );
      simBall.acc( target - playerpos );
      for( int i = 1; i <= 11; i++ )
	{
	  playerDist[i] = 0.0;
	  playerV[i] = 0.0;
	}
      numSteps = 0;
      score = 0;

      simBall.simulate();   
      // this is the ball after the kick command (assume players need time to realize that)
      while( fabs( simBall.pos.getX() ) < 0.5 * FIELD_LENGTH && numSteps < maxSimSteps )
	{
	  numSteps++;
	  simBall.simulate();

	  if( fabs( simBall.pos.getX() ) < 0.5 * FIELD_LENGTH )
	    {
	      for( int i = 1; i <= 11; i++ )
		if( simPlayer[i] )
		  {
		    const WPlayer & player = ourTeamGoalShot ? world->otherTeam[i] : world->ourTeam[i];
		    playerV[i] *= player.decay;

		    // player is immobile after tackling
		    if( player.cyclesSinceTackle + numSteps > TACKLE_CYCLES )
		      {
			playerV[i] += MAXPOWER * player.dashPowerRate;
			if( playerV[i] > player.vmax )
			  playerV[i] = player.vmax;
			playerDist[i] += playerV[i];

			if( player.isGoalie == GOALIE_TRUE )
			  {
			    if( player.distTo( simBall ) < playerDist[i] + CATCHABLE_AREA_L )
			      {
				score += 3;
				// very bad if opponent goalie does not have to move to get the ball
				if( player.distTo( simBall ) < CATCHABLE_AREA_L )
				  score += 3;
			      }
			  }
			else
			  {
			    if( player.distTo( simBall ) < playerDist[i] + player.kickRange() )
			      {
				score++;
				// very bad if opponent player does not have to move to get the ball
				if( player.distTo( simBall ) < player.kickRange() )
				  score++;
			      }
			  }
		      }
		  }
	    }
	}
      score += numSteps;

      // no goalshots near field border
      if( 2 * ( 0.5 * FIELD_LENGTH - fabs( playerpos.getX() ) ) < fabs( playerpos.getY() ) - 0.5 * GOAL_WIDTH )
	score += score_GOOD;      // makes this shot at most AVERAGE
      if( 0.5 * FIELD_LENGTH - fabs( playerpos.getX() ) < fabs( playerpos.getY() ) - 0.5 * GOAL_WIDTH )
	score += score_PERFECT;   // makes this shot at most GOOD

      if( score < minScore )
	{
	  minScore = score;
	  bestTarget = target;
	}
    }

  if( minScore < score_POOR )
    {
      grade.upgrade();
      if( minScore < score_AVERAGE )
	{
	  grade.upgrade();
	  if( minScore < score_GOOD )
	    {
	      grade.upgrade();
	      if( minScore < score_PERFECT )
		grade.upgrade();
	    }
	}
    }
  return grade;
};


//////////////////////////////////////////////////////////////////////
//
// GoalShotEval::correctCorner
//
//////////////////////////////////////////////////////////////////////

Vector2d GoalShotEval::correctCorner( double dist ) const
{
  // assume 10 percent error (this is what soccerserver may produce)
  return Vector2d( 0.0, max( minDistToCorner, min( maxDistToCorner, 0.1 * dist ) ) );
}
