//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "goalshotmodule.h"

#define MOD_GOAL

#include "Trace.h"

using namespace std;


//////////////////////////////////////////////////////////////////////
//
// class GoalShotModule
//
//////////////////////////////////////////////////////////////////////

GoalShotModule::GoalShotModule( Worldmodel * _world, TacticalModule * _tact,
				AdvSkills * _skills, GoalShotEval * _goalshoteval )
{
  world        = _world;
  tact         = _tact;
  skills       = _skills;
  goalshoteval = _goalshoteval;

  tackleToGoal = false;
  finishedKick = false;
  startedKick = false;
  availableKickTime = 0;
  goalTarget = Vector2d( 0.5 * FIELD_LENGTH, 0.0 );
};


//////////////////////////////////////////////////////////////////////
//
// bool GoalShotModule::feedbackWorld
//
//////////////////////////////////////////////////////////////////////

bool GoalShotModule::feedbackWorld()
{
  my = world->my;

  // does the player have to report a goalshot?
  if( finishedKick )
    {
      finishedKick = false;
      // did the player get visual information about the ball?
      if( world->ball.age == 0 )
	{
	  // will the ball stop behind the goalline?
	  if( ( world->ball.pos + world->ball.v * ( 1.0 / ( 1.0 - BALL_DECAY ) ) ).getX() >= 0.5 * FIELD_LENGTH )
	    {
	      // calculate the y coordinate where the ball will cross the goal line
	      double yCross = world->ball.pos.getY() +
		world->ball.v.getY() * ( 0.5 * FIELD_LENGTH - world->ball.pos.getX() ) / world->ball.v.getX();
	      // does the ball pass the goalline within the goal?
	      if ( fabs( yCross ) < 0.5 * GOAL_WIDTH ) 
		{
		  // report a goalshot
		  tact->reportOurTacticalAction( TA_GOALSHOT, my.number );
		}
	    }
	}
    }

  return true;
}; 
  

//////////////////////////////////////////////////////////////////////
//
// Grade GoalShotModule::evaluate
//
//////////////////////////////////////////////////////////////////////

Grade GoalShotModule::evaluate()
{
  Grade grade;

  tackleToGoal = false;

  // is player immobile from tackling?
  if( my.cyclesSinceTackle <= TACKLE_CYCLES )
    {
      grade.set_IMPOSSIBLE();
      return grade;
    }

  if( !my.ballKickable( world->ball ) )
    {
      // ball tackleable and not kickable?

      if ( ( world->ball.pos.getX() > 0.5 * FIELD_LENGTH - PENALTY_AREA_LENGTH - 3.0 ) &&
	   ( fabs( world->ball.pos.getY() ) < PENALTY_AREA_WIDTH ) &&
	   ( tact->firstDangerousOpponent != UNKNOWN ) &&
	   ( tact->firstResponsiblePlayer != UNKNOWN ) &&
	   ( tact->ourTeamReachBallTime[tact->firstResponsiblePlayer] 
	     >= tact->otherTeamReachBallTime[tact->firstDangerousOpponent] ) )
	{
	  tackleCommand = skills->tackle();
	  if( tackleCommand.failProb < .3 && tackleCommand.state != AdvSkills::TRANGE )
	    {
	      // y coordinate where the ball will cross the goal line
	      double yCross = world->ball.pos.getY() +
		tackleCommand.dir.getY() * ( 0.5 * FIELD_LENGTH - world->ball.pos.getX() ) / tackleCommand.dir.getX();

	      if( fabs( yCross ) < 0.5 * GOAL_WIDTH )
		{
		  grade.set_PERFECT();
		  tackleToGoal = true;
		  DECISION( "Goalshot impossible, tackling reasonable" );
		  return grade;
		}
	    }
	}

      grade.set_IMPOSSIBLE();
      return grade;
    }

  // now we are able to kick the ball 

  // how much time does the player have for a kick?
  availableKickTime = goalshoteval->maxKickDuration;
  if( tact->firstDangerousOpponent != UNKNOWN )
    availableKickTime = min( max( 1, tact->otherTeamReachBallTime[tact->firstDangerousOpponent] ),
			     availableKickTime );

  // if there might have been a player<->ball collision do not place the ball
  if( world->collState != NONE )
    availableKickTime = 1;

  grade = goalshoteval->teamEval( world, my.pos, goalTarget, true );

  return grade;
};


//////////////////////////////////////////////////////////////////////
//
// GoalShotModule::turnHeadEvaluate
//
//////////////////////////////////////////////////////////////////////

TurnHeadWish GoalShotModule::turnHeadEvaluate()
{
  TurnHeadWish thw;
  thw.lookNowhere( Grade::IMPOSSIBLE, VIEW_WIDTH_NARROW );

  // if opponent goal is old and ball is not too old look at it, if
  // ball is close to me and I am close to the opponent goal
  if( world->otherGoal.age > 2 &&
      // my.distTo( world->ball ) < 2 * VISIBLE_DISTANCE &&
      world->ball.age < 3 &&
      my.distTo( world->otherGoal ) < 25.0 )
    {
      // look at the middle of the opponent goal
      thw.lookAtOppGoal( Grade::PERFECT, VIEW_WIDTH_NARROW );
    }

  if( startedKick )
    {
      startedKick = false;
      thw.lookAtOppGoal( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
    }

  return thw;
}


//////////////////////////////////////////////////////////////////////
//
// GoalShotModule::act
//
//////////////////////////////////////////////////////////////////////

bool GoalShotModule::act()
{
  if( tackleToGoal )
    {
      skills->exec( tackleCommand );
      return true;
    }

  kickCommand = skills->kick( BALL_SPEED_MAX, goalTarget, availableKickTime );
  skills->exec( kickCommand );

  finishedKick = ( kickCommand.state == AdvSkills::KFINISHING );

  if( availableKickTime > 1 && kickCommand.state == AdvSkills::KPLACING )
    startedKick = true;

  return true; 
};  
