//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "tacticalmodule.h"

#include <string>
#include <sstream>

#define MOD_TACT

#include "Trace.h"

using namespace std;

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class TacticalSystem
//
///////////////////////////////////////////////////////////////////////////////////////////////

TacticalSystem::TacticalSystem( int _numDef, int _numDefMid, int _numOffMid, int _numOff )
{
  ensure( _numDef >= 0 && _numDefMid >= 0 && _numOffMid >= 0 && _numOff >= 0 );
  ensure( _numDef + _numDefMid + _numOffMid + _numOff == 10 );

  numDefenders   = _numDef;
  numDefMidfield = _numDefMid;
  numOffMidfield = _numOffMid;
  numMidfield    = _numDefMid + _numOffMid;
  numOffenders   = _numOff;
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class TacticalModule
//
///////////////////////////////////////////////////////////////////////////////////////////////

TacticalModule::TacticalModule( Worldmodel * _world, AdvSkills * _skills )
{
  world = _world;
  skills = _skills;

  ourBallControl = BC_UNKNOWN;

  ourSystem.numDefenders   = Config::getValueAsInt( "ourDefenders" );
  ourSystem.numDefMidfield = Config::getValueAsInt( "ourDefMidfield" );
  ourSystem.numOffMidfield = Config::getValueAsInt( "ourOffMidfield" );
  ourSystem.numMidfield    = ourSystem.numDefMidfield + ourSystem.numOffMidfield;
  ourSystem.numOffenders   = Config::getValueAsInt( "ourOffenders" );

  ourEdgeDefender      = Config::getValueAsInt( "tac_ourEdgeDefender" );
  ourEdgeDefMidfield   = Config::getValueAsInt( "tac_ourEdgeDefMidfield" );
  ourEdgeOffMidfield   = Config::getValueAsInt( "tac_ourEdgeOffMidfield" );
  ourEdgeOffender      = Config::getValueAsInt( "tac_ourEdgeOffender" );
  ourEdgeGoalie        = Config::getValueAsInt( "tac_ourEdgeGoalie" );

  otherEdgeDefender    = Config::getValueAsInt( "tac_otherEdgeDefender" );
  otherEdgeDefMidfield = Config::getValueAsInt( "tac_otherEdgeDefMidfield" );
  otherEdgeOffMidfield = Config::getValueAsInt( "tac_otherEdgeOffMidfield" );
  otherEdgeOffender    = Config::getValueAsInt( "tac_otherEdgeOffender" );
  otherEdgeGoalie      = Config::getValueAsInt( "tac_otherEdgeGoalie" );

  firstResponsiblePlayer  = UNKNOWN;
  secondResponsiblePlayer = UNKNOWN;
  thirdResponsiblePlayer  = UNKNOWN;

  firstDangerousOpponent  = UNKNOWN;
  secondDangerousOpponent = UNKNOWN;
  thirdDangerousOpponent  = UNKNOWN;

  for( int i = 1; i <= 11; i++ )
    {
      ourTeamReachBallTime[i]   = -1;
      otherTeamReachBallTime[i] = -1;

      // only do this for the coach (ToDo)
      playerTypeData[i] = new PlayerTypeData[playerParams.player_types];

      playerType[i] = UNKNOWN;

      playerPassInfo[i].grade.set_BAD();
      playerPassInfo[i].age = ETERNAL;
      playerPassInfo[i].xIndex = 0;
      playerPassInfo[i].yIndex = 0;
    }

  oldTime = -2;   // this should keep coach from doing bad things in timestep 0
  oldNumSubs = 0;

  attentionTo = UNKNOWN;
  cyclesSinceLastMessage = ETERNAL;

  communicateAction = false;
  communicateSimAction = false;
}


TacticalModule::~TacticalModule()
{
  for( int i = 1; i <= 11; i++ )
    delete [] playerTypeData[i];
}

bool
TacticalModule::reportOurTacticalAction( TacticalActionState state, const WBall &_simulatedBall,
					 int perPlayer, int recPlayer)
{
  communicateSimAction = true;
  simulatedBall = _simulatedBall;
  return reportOurTacticalAction( state, perPlayer, recPlayer );
};

bool
TacticalModule::reportOurTacticalAction( TacticalActionState state, int perPlayer, int recPlayer )
{
  bool newAction = false;

  switch( state )
    {
    case TA_CONTROL :
      if( perPlayer >= 1 && perPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = UNKNOWN;
	}
      break;

    case TA_PASS :
      if( perPlayer >= 1 && perPlayer <= 11 )
	// REMARK: have to handle emergency kicks sometime...
	if( recPlayer >= 1 && recPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = recPlayer;
	}
      break;
     
    case TA_GOALSHOT :
      if( perPlayer >= 1 && perPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = UNKNOWN;
	}
      break;

    case TA_REACT :
      newAction = true;
      reportedTAction.performer = ( perPlayer >= 1 && perPlayer <= 11 ? perPlayer : UNKNOWN );
      reportedTAction.receiver = UNKNOWN;
      break;

    case TA_UNKNOWN :
    default :
      break;
    }

  VALUE("reportOurTacticalAction-> newAction: " << newAction << " action:" << state \
	<< " perf: " << reportedTAction.performer << " recf: " << reportedTAction.receiver << endl);

  if( newAction )
    {
      reportedTAction.state = state;
      reportedTAction.quality = TQ_REPORTED;
      reportedTAction.age = 0;
      return true;
    }
	
  return false;
}


bool
TacticalModule::ourTeamDefender( int n ) const
{
  ensure( n >= 1 && n <= 11 );
  if( n < 1 || n > 11 )
    return false;
  int tn = world->ourTeam[n].tacticalNo;
  if( tn == UNKNOWN )
    return false;
  if( tn >= 2 && tn <= ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalModule::ourTeamDefMidfield( int n ) const
{
  ensure( n >= 1 && n <= 11 );
  if( n < 1 || n > 11 )
    return false;
  int tn = world->ourTeam[n].tacticalNo;
  if( tn == UNKNOWN )
    return false;
  if( tn > ourSystem.numDefenders + 1 &&
      tn <= ourSystem.numDefMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalModule::ourTeamOffMidfield( int n ) const
{
  ensure( n >= 1 && n <= 11 );
  if( n < 1 || n > 11 )
    return false;
  int tn = world->ourTeam[n].tacticalNo;
  if( tn == UNKNOWN )
    return false;
  if( tn > ourSystem.numDefMidfield + ourSystem.numDefenders + 1 &&
      tn <= ourSystem.numMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalModule::ourTeamMidfield( int n ) const
{
  ensure( n >= 1 && n <= 11 );
  if( n < 1 || n > 11 )
    return false;
  int tn = world->ourTeam[n].tacticalNo;
  if( tn == UNKNOWN )
    return false;
  if( tn > ourSystem.numDefenders + 1 &&
      tn <= ourSystem.numMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalModule::ourTeamOffender( int n ) const
{
  ensure( n >= 1 && n <= 11 );
  if( n < 1 || n > 11 )
    return false;
  int tn = world->ourTeam[n].tacticalNo;
  if( tn == UNKNOWN )
    return false;
  if( tn > ourSystem.numMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


void
TacticalModule::determineResponsibility()
{
  WBall     simBall = world->ball;  // simulated ball
  long      simTime = 0;            // number of simulated steps
  long      maxSimSteps = 100;      // maximum number of steps to simulate

  double    ourTeamStamina[12];     // guessed stamina used for simulation
  double    ourTeamDist[12];        // radius of players' movements
  double    ourTeamV[12];           // players' velocity
  bool      ourTeamSorted[12];      // has player "reached the ball"?
  int       ourTeamNumSim = 0;      // number of players that have "reached the ball"
  double    ourTeamRange[12];       // players' range (kickrange or catchrange)

  double    otherTeamStamina[12];   // same for other team
  double    otherTeamDist[12];
  double    otherTeamV[12];
  bool      otherTeamSorted[12];
  int       otherTeamNumSim = 0;
  double    otherTeamRange[12];

  bool      playerLeft = true;      // is there a player left to be sorted?

  firstResponsiblePlayer = UNKNOWN;
  secondResponsiblePlayer = UNKNOWN;
  thirdResponsiblePlayer = UNKNOWN;

  firstDangerousOpponent = UNKNOWN;
  secondDangerousOpponent = UNKNOWN;
  thirdDangerousOpponent = UNKNOWN;

  // if ball is close to a player we can not rely on its velocity so we set it to zero
  for( int i = 1; i <= 11; i++ )
    {
      if( world->ourTeam[i].distTo( simBall ) < world->ourTeam[i].kickRange() ||
	  world->otherTeam[i].distTo( simBall ) < world->otherTeam[i].kickRange() )
	{
	  simBall.v.setXY( 0.0, 0.0 );
	  break;
	}
    }

  for( int i = 1; i <= 11; i++ )
    {
      // use simulated stamina for team members
      ourTeamStamina[i] = max( 5 * MAXPOWER, world->ourTeam[i].stamina - EFFORT_DEC_THR * STAMINA_MAX );

      ourTeamSorted[i] = ( world->ourTeam[i].age > MAXAGE );
      ourTeamDist[i] = 0.0;
      ourTeamV[i] = 0.0;
      ourTeamReachBallTime[i] = ETERNAL;
      ourTeamReachBallPos[i] = world->ourTeam[i].pos;
      ourTeamRange[i] = ( world->ourTeam[i].isGoalie == GOALIE_TRUE ? CATCHABLE_AREA_L : world->ourTeam[i].kickRange() );

      otherTeamStamina[i] = STAMINA_MAX;   // world->otherTeam[i].stamina;   // (almost) worst case scenario

      otherTeamSorted[i] = ( world->otherTeam[i].age > MAXAGE );
      otherTeamDist[i] = 0.0;
      otherTeamV[i] = 0.0;
      otherTeamReachBallTime[i] = ETERNAL;
      otherTeamReachBallPos[i] = world->otherTeam[i].pos;
      otherTeamRange[i] = ( world->otherTeam[i].isGoalie == GOALIE_TRUE ? CATCHABLE_AREA_L : world->otherTeam[i].kickRange() );
    }

  // player's own stamina is well known
  ourTeamStamina[world->my.number] = world->my.stamina - EFFORT_DEC_THR * STAMINA_MAX;

  // let AdvSkills handle player himself (use controlBall to determine time player needs to reach the ball)

  // do not use skills if player is immobile from tackling
  bool useSkills = ( world->my.cyclesSinceTackle > TACKLE_CYCLES );

  long myReachTime = ETERNAL;
  AdvSkills::TurtleCommand turtleParams;
  WBall meet;

  // see what controlBall will report
  if( useSkills )
    {
      if( world->my.isGoalie == GOALIE_TRUE )
	{
	  turtleParams = skills->interceptTimed( world->my, simBall, world->my.stamina - EFFORT_DEC_THR * STAMINA_MAX,
						 CATCHABLE_AREA_L, meet);
	  if( !meet.inOurPenaltyArea() )
	    turtleParams = skills->interceptTimed( world->my, simBall, world->my.stamina - EFFORT_DEC_THR * STAMINA_MAX,
						   world->my.kickRange(), meet);
	}
      else
	turtleParams = skills->interceptTimed( world->my, simBall, world->my.stamina - EFFORT_DEC_THR * STAMINA_MAX,
					       world->my.kickRange(), meet);
      myReachTime = turtleParams.time;
    }

  while( playerLeft && simTime < maxSimSteps )
    {
      playerLeft = false;

      // simulate both teams
      for( int i = 1; i <= 11; i++ )
	{
	  // but not player himself
	  if( !useSkills || i != world->my.number )
	    {
	      // our player first
	      const WPlayer& ourPlayer = world->ourTeam[i];

	      if( !ourTeamSorted[i] )
		{
		  playerLeft = true;

		  // check if player is immobile from tackling (in this case he only gains stamina)
		  if( ourPlayer.cyclesSinceTackle <= TACKLE_CYCLES )
		    ourTeamStamina[i] = min( ourTeamStamina[i] + ourPlayer.staminaIncMax, STAMINA_MAX );
		  else
		    {
		      // has player already reached the ball?
		      if( dist( ourPlayer.pos, simBall.pos ) < ourTeamDist[i] + ourTeamRange[i] )
			{
			  ourTeamSorted[i] = true;
			  ourTeamReachBallTime[i] = simTime;
			  ourTeamReachBallPos[i] = simBall.pos;
			  switch( ourTeamNumSim )
			    {
			    case 0  : firstResponsiblePlayer = i; break;
			    case 1  : secondResponsiblePlayer = i; break;
			    case 2  : thirdResponsiblePlayer = i; break;
			    default : break;
			    }
			  ourTeamNumSim++;
			}
		      else   // if not: simulate player's movement
			{
			  ourTeamV[i] = min( ourPlayer.vmax,
					     ( ourPlayer.staminaIncMax +
					       max( 0.0, min( MAXPOWER - ourPlayer.staminaIncMax, ourTeamStamina[i] ) ) )
					     * ourPlayer.dashPowerRate + ourTeamV[i] * ourPlayer.decay );

			  ourTeamStamina[i] -= max( 0.0, min( MAXPOWER - ourPlayer.staminaIncMax, ourTeamStamina[i] ) );
			  ourTeamStamina[i] = max( 0.0, ourTeamStamina[i] );
			  ourTeamDist[i] += ourTeamV[i];
			}
		    }
		}
	    }

	  // now opponent player
	  const WPlayer& otherPlayer = world->otherTeam[i];

	  if( !otherTeamSorted[i] )
	    {
	      playerLeft = true;
	      // has player already reached the ball?
	      if( dist( otherPlayer.pos, simBall.pos ) < otherTeamDist[i] + otherTeamRange[i] )
		{
		  otherTeamSorted[i] = true;
		  otherTeamReachBallTime[i] = simTime;
		  otherTeamReachBallPos[i] = simBall.pos;
		  switch( otherTeamNumSim )
		    {
		    case 0  : firstDangerousOpponent = i; break;
		    case 1  : secondDangerousOpponent = i; break;
		    case 2  : thirdDangerousOpponent = i; break;
		    default : break;
		    }
		  otherTeamNumSim++;
		}
	      else   // if not: simulate player's movement
		{
		  otherTeamV[i] = min( otherPlayer.vmax,
				       ( otherPlayer.staminaIncMax +
					 max( 0.0, min( MAXPOWER - otherPlayer.staminaIncMax, otherTeamStamina[i] ) ) )
				       * otherPlayer.dashPowerRate + otherTeamV[i] * otherPlayer.decay );

		  otherTeamStamina[i] -= max( 0.0, min( MAXPOWER - otherPlayer.staminaIncMax, otherTeamStamina[i] ) );
		  otherTeamStamina[i] = max( 0.0, otherTeamStamina[i] );
		  otherTeamDist[i] += otherTeamV[i];
		}
	    }
	}

      simBall.simulate();
      simTime++;
    }

  // if player was simulated by AdvSkills sort him into responsible players
  if( useSkills )
    {
      if( firstResponsiblePlayer == UNKNOWN || myReachTime <= ourTeamReachBallTime[firstResponsiblePlayer] )
	{
	  thirdResponsiblePlayer = secondResponsiblePlayer;
	  secondResponsiblePlayer = firstResponsiblePlayer;
	  firstResponsiblePlayer = world->my.number;
	  ourTeamReachBallTime[world->my.number] = myReachTime;
	  ourTeamReachBallPos[world->my.number] = meet.pos;
	}
      else if( secondResponsiblePlayer == UNKNOWN || myReachTime <= ourTeamReachBallTime[secondResponsiblePlayer] )
	{
	  thirdResponsiblePlayer = secondResponsiblePlayer;
	  secondResponsiblePlayer = world->my.number;
	  ourTeamReachBallTime[world->my.number] = myReachTime;
	  ourTeamReachBallPos[world->my.number] = meet.pos;
	}
      else if( thirdResponsiblePlayer == UNKNOWN || myReachTime <= ourTeamReachBallTime[thirdResponsiblePlayer] )
	{
	  thirdResponsiblePlayer = world->my.number;
	  ourTeamReachBallTime[world->my.number] = myReachTime;
	  ourTeamReachBallPos[world->my.number] = meet.pos;
	}
    }
}


void
TacticalModule::determineBallControl()
{
  BallControlState ourBallControlNew = BC_UNKNOWN;     // new value for ourBallControl

  long firstFriendReachTime = -1;     // time when firstResponsiblePlayer will reach the ball
  long firstOpponentReachTime = -1;   // time when firstDangerousOpponent will reach the ball

  switch( world->playMode )
    {
    case PLAY_MODE_UNKNOWN :
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
    case PLAY_MODE_PAUSE :
    case PLAY_MODE_PENALTY_TAKEN_US :
    case PLAY_MODE_PENALTY_TAKEN_THEM :
    case PLAY_MODE_PENALTY_MISS_US :
    case PLAY_MODE_PENALTY_MISS_THEM :
    case PLAY_MODE_PENALTY_SCORE_US :
    case PLAY_MODE_PENALTY_SCORE_THEM :

      // there is no game running
      ourBallControlNew = BC_UNKNOWN;
      break;

    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_GOAL_KICK_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_GOALIE_CATCH_US :
    case PLAY_MODE_BACKPASS_THEM :
    case PLAY_MODE_FREEKICK_FAULT_THEM :
    case PLAY_MODE_FOUL_CHARGE_THEM :
    case PLAY_MODE_FOUL_PUSH_THEM :
    case PLAY_MODE_FOUL_MULTIPLE_ATTACK_THEM :
    case PLAY_MODE_FOUL_BALLOUT_THEM :
    case PLAY_MODE_PENALTY_SETUP_US :
    case PLAY_MODE_PENALTY_READY_US :

      // in a standard situation for us ball control is obvious
      ourBallControlNew = BC_SAFE;
      break;

    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_OFFSIDE_US :
    case PLAY_MODE_GOALIE_CATCH_THEM :
    case PLAY_MODE_BACKPASS_US :
    case PLAY_MODE_FREEKICK_FAULT_US :
    case PLAY_MODE_FOUL_CHARGE_US :
    case PLAY_MODE_FOUL_PUSH_US :
    case PLAY_MODE_FOUL_MULTIPLE_ATTACK_US :
    case PLAY_MODE_FOUL_BALLOUT_US :
    case PLAY_MODE_PENALTY_SETUP_THEM :
    case PLAY_MODE_PENALTY_READY_THEM :

      // in a standard situation for the opponent team ball control is obvious
      ourBallControlNew = BC_LOST;
      break;

    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_PLAY_ON :
    case PLAY_MODE_HUMAN_JUDGE :

      // is there a player of our team? (there should be at least one...)
      if( firstResponsiblePlayer != UNKNOWN )
	{
	  firstFriendReachTime = ourTeamReachBallTime[firstResponsiblePlayer];

	  int ourEdge = ( ourTeamDefender( firstResponsiblePlayer ) ? ourEdgeDefender :
			  ourTeamDefMidfield( firstResponsiblePlayer ) ? ourEdgeDefMidfield :
			  ourTeamOffMidfield( firstResponsiblePlayer ) ? ourEdgeOffMidfield :
			  ourTeamOffender( firstResponsiblePlayer ) ? ourEdgeOffender :
			  ourEdgeGoalie );

	  int otherEdge = ( ourTeamDefender( firstResponsiblePlayer ) ? otherEdgeDefender :
			    ourTeamDefMidfield( firstResponsiblePlayer ) ? otherEdgeDefMidfield :
			    ourTeamOffMidfield( firstResponsiblePlayer ) ? otherEdgeOffMidfield :
			    ourTeamOffender( firstResponsiblePlayer ) ? otherEdgeOffender :
			    ourEdgeGoalie );

	  // is there a player of the opponent team?
	  if( firstDangerousOpponent != UNKNOWN )
	    {
	      firstOpponentReachTime = otherTeamReachBallTime[firstDangerousOpponent];

	      // opponent is faster
	      if( firstOpponentReachTime + otherEdge < firstFriendReachTime )
		{
		  ourBallControlNew = ( ourBallControl == BC_SAFE ? BC_DANGER : BC_LOST );
		}
	      // we are faster
 	      else if( firstFriendReachTime + ourEdge < firstOpponentReachTime )
		{
		  ourBallControlNew = ( ourBallControl == BC_LOST ? BC_DANGER : BC_SAFE );
		}
	      // dangerous situation
	      else
		{
		  ourBallControlNew = BC_DANGER;
		}
	    }
	  else   // no opponent seen
	    {
	      ourBallControlNew = ( ourBallControl == BC_LOST ? BC_DANGER : BC_SAFE);
	    }
	}
      else   // weird case: no player of our team found...
	{
	  // is there at least an opponent?
	  if( firstDangerousOpponent != UNKNOWN )
	    {
	      firstOpponentReachTime = otherTeamReachBallTime[firstDangerousOpponent];
	  
	      ourBallControlNew = ( ourBallControl == BC_SAFE ? BC_DANGER : BC_LOST);
	    }
	  else   // there is NOBODY around... try to use friend's ball control values
	    {
	      ourBallControlNew = BC_UNKNOWN;
	    }
	}
      break;

    default :

      ensure( 0 );
      break;
    }

  ourBallControl = ourBallControlNew;
}


bool
TacticalModule::determineOurTacticalAction()
{
  bool useHistory = false;     // are tactical history and world consistent?

  // increase age for our action
  if( ourAction.age < ETERNAL )
    ourAction.age++;

  TacticalAction ourActionNew( TA_UNKNOWN, UNKNOWN, UNKNOWN, 0 );

  switch( world->playMode )
    {
    case PLAY_MODE_UNKNOWN :
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
    case PLAY_MODE_PAUSE :
      
      ourActionNew.quality = TQ_GUESSED;
      break;

    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_GOAL_KICK_US :   // this is not a "standard standard situation"!!!
      // REMARK: in a standard situation for us the state can only be TA_CONTROL or
      // TA_PASS, and the latter only in PLAY_MODE_GOAL_KICK_US
    case PLAY_MODE_GOALIE_CATCH_US :
    case PLAY_MODE_BACKPASS_THEM :
    case PLAY_MODE_FREEKICK_FAULT_THEM :
    case PLAY_MODE_FOUL_CHARGE_THEM :
    case PLAY_MODE_FOUL_PUSH_THEM :
    case PLAY_MODE_FOUL_MULTIPLE_ATTACK_THEM :
    case PLAY_MODE_FOUL_BALLOUT_THEM :

      // ---< 1 >--- take reported information into account
      if( reportedTAction.quality == TQ_REPORTED &&
	  ( reportedTAction.state == TA_CONTROL ||
	    world->playMode == PLAY_MODE_GOAL_KICK_US && reportedTAction.state == TA_PASS ) &&
	  reportedTAction.performer != UNKNOWN )
	{
	  ourActionNew.state     = reportedTAction.state;
	  ourActionNew.quality   = TQ_REPORTED;
	  ourActionNew.performer = reportedTAction.performer;
	  ourActionNew.receiver  = reportedTAction.receiver;
	  break;
	}

      // ---< 2 >--- take audial information into account
      if( ( audioTAction.state == TA_CONTROL ||
	    ( world->playMode == PLAY_MODE_GOAL_KICK_US && audioTAction.state == TA_PASS ) ) &&
	  audioTAction.performer != UNKNOWN )
	{
	  // if sending player is sure about what he sent... believe him
	  if( audioTAction.quality == TQ_REPORTED )
	    {
	      ourActionNew.state     = audioTAction.state;
	      ourActionNew.quality   = TQ_AUDIAL_REPORTED;
	      ourActionNew.performer = audioTAction.performer;
	      ourActionNew.receiver  = audioTAction.receiver;
	      break;
	    }
	  // if he is not so sure compare it with our results
	  if( audioTAction.quality == TQ_AUDIAL_REPORTED )
	    {
	      if( audioTAction.performer == firstResponsiblePlayer ||
		  audioTAction.performer == secondResponsiblePlayer ||
		  audioTAction.performer == thirdResponsiblePlayer )
		{
		  ourActionNew.state     = audioTAction.state;
		  ourActionNew.quality   = TQ_AUDIAL;
		  ourActionNew.performer = audioTAction.performer;
		  ourActionNew.receiver  = audioTAction.receiver;
		  break;
		}
	    }
	}

      // ---< 3 >--- try to determine performer by history

      // first case : goal kick situation and pass action
      if( world->playMode == PLAY_MODE_GOAL_KICK_US && ourAction.state == TA_PASS &&
	  ( ourAction.quality == TQ_REPORTED || ourAction.quality == TQ_AUDIAL_REPORTED ) &&
	  ourAction.performer != UNKNOWN )
	{
	  // keep that until some player changes it (to control)
	  ourActionNew = ourAction;
	  break;
	}
	  
      // second case : control action
      if( ourAction.state     == TA_CONTROL &&
	  ourAction.quality   != TQ_UNKNOWN &&
	  ourAction.performer != UNKNOWN )
	{
	  // if old performer fits into new worldmodel or is a reported one keep him
	  if( ourAction.performer == firstResponsiblePlayer ||
	      ourAction.performer == secondResponsiblePlayer ||
	      ourAction.performer == thirdResponsiblePlayer ||
	      ourAction.quality   == TQ_REPORTED ||
	      ourAction.quality   == TQ_AUDIAL_REPORTED )
	    {
	      ourActionNew = ourAction;
	      break;
	    }
	}

      // ---< 4 >--- if all else fails, take firstResponsiblePlayer as performer of CONTROL
      ourActionNew.state     = TA_CONTROL;
      ourActionNew.quality   = TQ_GUESSED;
      ourActionNew.performer = firstResponsiblePlayer;
      ourActionNew.receiver  = UNKNOWN;
      break;

    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_OFFSIDE_US :
    case PLAY_MODE_GOALIE_CATCH_THEM :
    case PLAY_MODE_BACKPASS_US :
    case PLAY_MODE_FREEKICK_FAULT_US :
    case PLAY_MODE_FOUL_CHARGE_US :
    case PLAY_MODE_FOUL_PUSH_US :
    case PLAY_MODE_FOUL_MULTIPLE_ATTACK_US :
    case PLAY_MODE_FOUL_BALLOUT_US :

      // in a standard situation for the opponent team our tactical action is obvious
      ourActionNew.state   = TA_REACT;
      ourActionNew.quality = TQ_GUESSED;
      break;

      case PLAY_MODE_PENALTY_SETUP_US:
      case PLAY_MODE_PENALTY_SETUP_THEM:
      case PLAY_MODE_PENALTY_READY_US:
      case PLAY_MODE_PENALTY_READY_THEM:
      case PLAY_MODE_PENALTY_TAKEN_US:
      case PLAY_MODE_PENALTY_TAKEN_THEM:
      case PLAY_MODE_PENALTY_MISS_US:
      case PLAY_MODE_PENALTY_MISS_THEM:
      case PLAY_MODE_PENALTY_SCORE_US:
      case PLAY_MODE_PENALTY_SCORE_THEM:

    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_PLAY_ON :
    case PLAY_MODE_HUMAN_JUDGE :

      // step 1 : first take reported information into account
      if( reportedTAction.quality == TQ_REPORTED && reportedTAction.state != TA_UNKNOWN )
	{
	  ourActionNew.quality   = TQ_REPORTED;
	  ourActionNew.state     = reportedTAction.state;
	  ourActionNew.performer = reportedTAction.performer;
	  ourActionNew.receiver  = reportedTAction.receiver;
	  break;
	}

      // step 2 : if there is no reported information take audio information into account
      if( audioTAction.state != TA_UNKNOWN )
	if( audioTAction.quality == TQ_REPORTED || audioTAction.quality == TQ_AUDIAL_REPORTED )
	  // do not overwrite better information
	  if( ( ourAction.quality != TQ_REPORTED && ourAction.quality != TQ_AUDIAL_REPORTED ) ||
	      audioTAction.age < ourAction.age )
	    // do not use information that mentions myself as performer - I should know what I do!!!
	    if( audioTAction.performer != world->my.number || ourAction.state == TA_UNKNOWN )
	      {
		ourActionNew.quality   = ( audioTAction.quality == TQ_REPORTED ? TQ_AUDIAL_REPORTED : TQ_AUDIAL );
		ourActionNew.state     = audioTAction.state;
		ourActionNew.performer = audioTAction.performer;
		ourActionNew.receiver  = audioTAction.receiver;
		// use our team mate's starting time for that action
		ourActionNew.age       = audioTAction.age;
		break;
	      }

      // step 3 : if there is no good audio information try to use history
      if( ourAction.quality != TQ_UNKNOWN )
	{
	  switch( ourAction.state )
	    {
	    case TA_CONTROL :
	      // keep old action if ball is not lost
	      // and if performer is still responsible
	      if( ourBallControl != BC_LOST &&
		  ( ourAction.performer == firstResponsiblePlayer ||
		    ourAction.performer == secondResponsiblePlayer ||
		    ourAction.performer == thirdResponsiblePlayer ) )
		// before: || ourAction.quality == TQ_REPORTED ||
		// ourAction.quality == TQ_AUDIAL_REPORTED ) )
		// does not make sense in non-standard situations
		{
		  useHistory = true;
		}
	      break;
	      
	    case TA_PASS :
	      // keep old action if ball is not lost...
	      if( ( ourBallControl != BC_LOST ||
		    firstDangerousOpponent == UNKNOWN ||
		    otherTeamReachBallTime[firstDangerousOpponent] > 5 ) &&
		  // ...and if no friend has it under control
		  (
		    firstResponsiblePlayer == ourAction.performer ||
		    firstResponsiblePlayer == UNKNOWN ||
		    ourTeamReachBallTime[firstResponsiblePlayer] > 2 ) )
		{
		  useHistory = true;
		}
	      break;

	    case TA_GOALSHOT :
	      // keep old action if ball is not lost
	      if( ourBallControl != BC_LOST ||
		  firstDangerousOpponent == UNKNOWN ||
		  otherTeamReachBallTime[firstDangerousOpponent] > 2 )
		{
		  useHistory = true;
		}
	      break;

	    case TA_REACT :
	      // keep old action if ball control is not safe again
	      if( ourBallControl != BC_SAFE )
		{
		  useHistory = true;
		}
	      break;

	    case TA_UNKNOWN :
	    default :
	      break;
	    }
	}
      
      // use history if it is consistent with world and not too old
      if( useHistory &&
	  ( ourAction.state == TA_CONTROL  ? ourAction.age < 20 :
	    ( ourAction.state == TA_PASS     ? ourAction.age < 10 :
	      ( ourAction.state == TA_GOALSHOT ? ourAction.age < 15 :   // 10 ?
		( ourAction.state == TA_REACT    ? ourAction.age < 20 : false ) ) ) ) )
	{
	  ourActionNew = ourAction;
	  break;
	}

      // history and world are not consistent -> try to determine tactical action
      // by looking at the world

      switch( ourBallControl )
	{
	case BC_SAFE :
	  if( firstResponsiblePlayer == UNKNOWN )
	    {
	      ourActionNew.state = TA_REACT;
	    }
	  else
	    {
	      ourActionNew.state = TA_CONTROL;
	      ourActionNew.performer = firstResponsiblePlayer;
	    }
	  break;

	case BC_DANGER :
	  // use history in that case
	  if( ourAction.state == TA_REACT || firstResponsiblePlayer == UNKNOWN )
	    {
	      ourActionNew.state = TA_REACT;
	      ourActionNew.performer = firstResponsiblePlayer;
	    }
	  else
	    {
	      ourActionNew.state = TA_CONTROL;
	      ourActionNew.performer = firstResponsiblePlayer;
	    }
	  break;

	case BC_LOST :
	  ourActionNew.state = TA_REACT;
	  ourActionNew.performer = firstResponsiblePlayer;
	  break;

	case BC_UNKNOWN :
	default :
	  ourActionNew.state = TA_REACT;
	  ourActionNew.performer = firstResponsiblePlayer;
	  break;
	}

      ourActionNew.quality = TQ_GUESSED;
      break;

    default :

      ensure( 0 );
      break;
    }

  // if old and new action for our team are the same, don't change age
  // (only if ourActionNew comes from audial information)
  if( ourAction.state == ourActionNew.state &&
      ourAction.performer == ourActionNew.performer &&
      ourAction.receiver == ourActionNew.receiver )
    {
      ourActionNew.age = max( ourAction.age, ourActionNew.age );
    }

  ourAction = ourActionNew;

  // if action is very old, downgrade its quality
  if( ourAction.age > 10 )
    {
      if( ourAction.quality == TQ_REPORTED )
	ourAction.quality = TQ_GUESSED;
      if( ourAction.quality == TQ_AUDIAL_REPORTED )
	ourAction.quality = TQ_AUDIAL;
    }

  return true;
}


void
TacticalModule::determineOtherPlayerTypes()
{
  int numTypes = playerParams.player_types;
  bool playerSubstituted;

  // only evaluate in PLAY_ON mode and if we had a worldmodel one timestep ago
  if( world->playMode == PLAY_MODE_PLAY_ON && world->time - oldTime == 1 )
    {
      for( int i = 1; i <= 11; i++ )
	{
	  playerSubstituted = false;
	  if( world->numOtherSubs > oldNumSubs )
	    {
	      for( int k = 0; k < world->numOtherSubs - oldNumSubs; k++ )
		{
		  if( world->otherSubs[( world->numOtherSubs - 1 - k ) %
				      ( playerParams.subs_max + 11 )].number == i )
		    {
		      VALUE( "Opp. player " << i << " was substituted!" );
		      playerSubstituted = true;
		      playerType[i] = UNKNOWN;

		      for( int j = 0; j < numTypes; j++ )
			{
			  playerTypeData[i][j].state = PTS_UNKNOWN;
			  playerTypeData[i][j].score = 0;
			}
		    }
		}
	    }

	  if( playerSubstituted )
	    continue;
	  if( playerType[i] != UNKNOWN )
	    continue;

	  const WPlayer& newPlayer = world->otherTeam[i];
	  double posDiff = norm( newPlayer.pos - oldPlayer[i].pos );
	  for( int j = 0; j < numTypes; j++ )
	    if( playerTypeData[i][j].state == PTS_UNKNOWN )
	      {
		if( posDiff > world->playerType[j].playerSpeedMax )
		  {
		    DECISION( "excluded type " << j << " for opp. player " << i );
		    playerTypeData[i][j].state = PTS_IMPOSSIBLE;
		    playerTypeData[i][j].score = 0;
		    continue;
		  }

		if( ( fabs( MAXMOMENT ) /
		      fabs( newPlayer.bodyDir - oldPlayer[i].bodyDir ) - 1 )
		    / norm( oldPlayer[i].v ) < world->playerType[j].inertiaFactor )
		  {
		    DECISION( "excluded type " << j << " for opp. player " << i );
		    playerTypeData[i][j].state = PTS_IMPOSSIBLE;
		    playerTypeData[i][j].score = 0;
		    continue;
		  }

		Vector2d accel = newPlayer.pos - oldPlayer[i].pos - oldPlayer[i].v;
		if( norm( accel ) / max( MINPOWER, MAXPOWER ) > world->playerType[j].dashPowerRate )
		  {
		    DECISION( "excluded type " << j << " for opp. player " << i );
		    playerTypeData[i][j].state = PTS_IMPOSSIBLE;
		    playerTypeData[i][j].score = 0;
		    continue;
		  }

		double decay = norm( oldPlayer[i].v + accel ) / norm( newPlayer.v );
		if( decay > world->playerType[j].playerDecay )
		  {
		    DECISION( "excluded type " << j << " for opp. player " << i );
		    playerTypeData[i][j].state = PTS_IMPOSSIBLE;
		    playerTypeData[i][j].score = 0;
		    continue;
		  }
	      }

	  if( playerType[i] == UNKNOWN )
	    {
	      int lastNotImpossible = -1;
	      int numImpossible = 0;
	      // check if all types but one are excluded
	      for( int j = 0; j < numTypes; j++ )
		{
		  if( playerTypeData[i][j].state == PTS_IMPOSSIBLE )
		    numImpossible++;
		  else
		    lastNotImpossible = j;
		}
	      if( numImpossible == numTypes - 1 )
		{
		  playerTypeData[i][lastNotImpossible].state = PTS_SURE;
		  playerType[i] = lastNotImpossible;
		  DECISION( "Opp. player " << i << " has type " << lastNotImpossible );
		}
	    }
	}
    }

  // finally reset oldPlayer array and oldTime
  for( int i = 1; i <= 11; i++ )
    oldPlayer[i] = world->otherTeam[i];
  oldTime = world->time;
  oldNumSubs = world->numOtherSubs;
}


AdvSkills::TurnHeadCommand
TacticalModule::calcTurnHead( TurnHeadWish goalTHW, TurnHeadWish passTHW, TurnHeadWish ballhTHW,
			      TurnHeadWish posTHW, TurnHeadWish stdTHW )
{
  // avoid undefined grades (set to IMPOSSIBLE)
  if( !goalTHW.grade.is_defined() )
    {
      goalTHW.grade.set_IMPOSSIBLE();
      cerr << "TACT: " << world->time << ", no " << world->my.number << ": goalTHW.grade undefined!" << endl;
    }
  if( !passTHW.grade.is_defined() )
    {
      passTHW.grade.set_IMPOSSIBLE();
      cerr << "TACT: " << world->time << ", no " << world->my.number << ": passTHW.grade undefined!" << endl;
    }
  if( !ballhTHW.grade.is_defined() )
    {
      ballhTHW.grade.set_IMPOSSIBLE();
      cerr << "TACT: " << world->time << ", no " << world->my.number << ": ballhTHW.grade undefined!" << endl;
    }
  if( !posTHW.grade.is_defined() )
    {
      posTHW.grade.set_IMPOSSIBLE();
      cerr << "TACT: " << world->time << ", no " << world->my.number << ": posTHW.grade undefined!" << endl;
    }
  if( !stdTHW.grade.is_defined() )
    {
      stdTHW.grade.set_IMPOSSIBLE();
      cerr << "TACT: " << world->time << ", no " << world->my.number << ": stdTHW.grade undefined!" << endl;
    }

  VALUE( "turnHead grades: std="<< stdTHW.grade << ", goal=" << goalTHW.grade << ", pass="
	 << passTHW.grade << ", ballh=" << ballhTHW.grade << ", pos=" << posTHW.grade );

  // sort the turn head wishes according to grade
  TurnHeadWish myTHW[5];

  myTHW[0] = goalTHW;
  myTHW[1] = passTHW;
  myTHW[2] = ballhTHW;
  myTHW[3] = posTHW;
  myTHW[4] = stdTHW;

  bool swapped = true;
  TurnHeadWish dummyTHW;

  // dumb bubble sort
  while( swapped )
    {
      swapped = false;
      for( int i = 4; i >= 1; i-- )
	if( myTHW[i].grade.is_better_than( myTHW[i-1].grade ) )
	  {
	    swapped = true;
	    dummyTHW = myTHW[i-1];
	    myTHW[i-1] = myTHW[i];
	    myTHW[i] = dummyTHW;
	  }
    }

  // body dir in the next cycle
  Angle myNextBodyDir = world->my.bodyDir + Angle( skills->getTurnHeadTurn() / world->my.inertiaMoment() );

  AdvSkills::TurnHeadCommand myTHC;

  view_width vWidth = world->viewWidth;   // keep view width unchanged if no module wants to change it

  int index = 0;
  bool leave = false;
  bool useAudio = false;

  while( !leave && index <= 4 && myTHW[index].grade.is_better_than( Grade::BAD ) )
    {
      switch( myTHW[index].type )
	{
	case TH_BALL:
	  myTHW[index].pos = world->ball.pos + world->ball.v;
	  if( ( firstDangerousOpponent != UNKNOWN &&
		otherTeamReachBallTime[firstDangerousOpponent] <= 1 ) ||
	      ( firstResponsiblePlayer != UNKNOWN &&
		ourTeamReachBallTime[firstResponsiblePlayer] <= 1 ) )
	    myTHW[index].pos = world->ball.pos;
	  break;

	case TH_OURTEAM:
	  myTHW[index].pos = world->ourTeam[myTHW[index].number].pos;
	  break;

	case TH_OPPTEAM:
	  myTHW[index].pos = world->otherTeam[myTHW[index].number].pos;
	  break;

	case TH_OURGOAL:
	  myTHW[index].pos = world->ourGoal.pos;
	  break;

	case TH_OPPGOAL:
	  myTHW[index].pos = world->otherGoal.pos;
	  break;

	case TH_POSITION:
	  // position is already set
	  break;

	case TH_BODYDIR:
	  myTHC = skills->turnHead( myNextBodyDir );
	  myTHC.vwidth = myTHW[index].vWidth;
	  return myTHC;

	case TH_LOOKAUDIO:
	  vWidth = myTHW[index].vWidth;
	  useAudio = true;
	  leave = true;
	  break;

	case TH_LOOKAROUND:
	  vWidth = myTHW[index].vWidth;
	case TH_UNKNOWN:
	default:
	  leave = true;
	  break;
	}

      // check if turn head wish makes sense
      if( !leave )
	{
	  double thisvAngle = world->getViewAngle( myTHW[index].vWidth );

	  if( fabs( ( myTHW[index].pos - world->my.pos ).getAngle() - myNextBodyDir ) <=
	      fabs( DEG_TO_RAD( MAXNECKANG ) ) + thisvAngle )
	    {
	      myTHC = skills->turnHead( myTHW[index].pos );
	      myTHC.vwidth = myTHW[index].vWidth;
	      DECISION("Turning head to: " << 
		       ( myTHW[index].type == TH_BALL ? "ball" :
			 myTHW[index].type == TH_OURTEAM ? "our Team" :
			 myTHW[index].type == TH_OPPTEAM ? "opp Team" :
			 myTHW[index].type == TH_OURGOAL ? "our goal" :
			 myTHW[index].type == TH_OPPGOAL ? "opp goal" :
			 myTHW[index].type == TH_POSITION ? "position" :
			 myTHW[index].type == TH_LOOKAUDIO ? "audio info" :
			 myTHW[index].type == TH_LOOKAROUND ? "around" 
			 : "unknown" )
		       << " pos: " << myTHW[index].pos
		       << " no: " << myTHW[index].number 
		       << " vwidth: " << 
		       ( myTHW[index].vWidth == VIEW_WIDTH_NARROW ? " narrow" :
			 myTHW[index].vWidth == VIEW_WIDTH_NORMAL ? " normal" :
			 myTHW[index].vWidth == VIEW_WIDTH_WIDE ? " wide" :
			 "unknown" )
		       );
	      return myTHC;
	    }
	  index++;   // try next turn head wish
	}
    }

  // if we reach this point none of the turn head wishes made sense
  // or a module wished to just look around
  // now we use the vDir, vAngle and vAge information stored in the worldmodel
  // to determine a good direction to look into

  // [MW] first, check if we have heard players which we haven't seen
  if( useAudio )
    {
      DECISION( "trying to use audio directions" );
      double checkDir = myNextBodyDir - DEG_TO_RAD( MAXNECKANG );
      double width = world->getViewAngle( vWidth );
      int hits;
      int bestHits = 0;
      double bestDir = 0.0;

      while ( fabs( checkDir - myNextBodyDir ) <= DEG_TO_RAD( MAXNECKANG + 1 ) )
	{
	  hits = 0;
	  for ( int i = 0; i < VISUALSTEPS ; i ++ )
	    {
	      if( fabs( checkDir - world->aOppDir[i] ) < width &&
		  world->dirAge( world->aOppDir[i] ) > world->aOppAge[i] )
		hits+=2;
	      if( fabs( checkDir - world->aOurDir[i] ) < width &&
		  world->dirAge( world->aOurDir[i] ) > world->aOurAge[i] )
		hits++;
	    }
	  if( hits > bestHits || 
	      ( hits == bestHits && world->dirAge( checkDir ) > world->dirAge( bestDir ) ) )
	    {
	      bestHits = hits;
	      bestDir = checkDir;
	    }
	  checkDir += 2 * width;
	}
      if( bestHits != 0 )
	{	  
	  VALUE( "found " << bestHits << " hits" );
	  myTHC = skills->turnHead( bestDir );
	  myTHC.vwidth = vWidth;
	  return myTHC;
	}
    }

  DECISION("Turning head to look around "); 

  const int NUM_DIRECTIONS = 72;
  double dirScore[NUM_DIRECTIONS];
  for( int i = 0; i < NUM_DIRECTIONS; i++ )
    dirScore[i] = 0;

  // if we received no visual info this cycle until now, don't worry.
  // if we use VIEW_WIDTH_NARROW the information will still reach us
  if( world->viewWidth == VIEW_WIDTH_NARROW )
    if( world->visualAge > 0 )
      {
	Angle curViewAngle = fabs( world->getViewAngle( world->viewWidth ) );
	for( int i = 0; i < NUM_DIRECTIONS; i++ )
	  if( fabs( DEG_TO_RAD( i * 360.0 / NUM_DIRECTIONS ) - world->my.faceDir ) < curViewAngle )
	    dirScore[i] += 1.0;
      }

  for( int j = 0; j < VISUALSTEPS; j++ )
    for( int i = 0; i < NUM_DIRECTIONS; i++ )
      if( fabs( DEG_TO_RAD( i * 360.0 / NUM_DIRECTIONS ) - world->vDir[j] ) < fabs( world->vAngle[j] ) )
	dirScore[i] += 1.0 / ( world->vAge[j] + 1 );

  double minScore = NUM_DIRECTIONS * ( VISUALSTEPS + 1 ) + 1.0;
  Angle minDir = myNextBodyDir;

  double thisScore;
  Angle thisDir;

  double vAngle = world->getViewAngle( vWidth );

  for( int i = 0; i < NUM_DIRECTIONS; i++ )
    {
      thisDir = DEG_TO_RAD( i * 360.0 / NUM_DIRECTIONS );
      thisScore = 0.0;
      // is it possible to look into that direction?
      if( fabs( thisDir - myNextBodyDir ) <= fabs( DEG_TO_RAD( MAXNECKANG ) ) )
	{
	  for( int j = 0; j < NUM_DIRECTIONS; j++ )
	    if( fabs( DEG_TO_RAD( j * 360.0 / NUM_DIRECTIONS ) - thisDir ) < vAngle )
	      thisScore += dirScore[j];
	  if( thisScore < minScore )
	    {
	      minScore = thisScore;
	      minDir = thisDir;
	    }
	}
    }

  myTHC = skills->turnHead( minDir );
  myTHC.vwidth = vWidth;
  return myTHC;
}


bool
TacticalModule::feedbackWorld()
{
  communicateSimAction = false;
  if( cyclesSinceLastMessage < ETERNAL )
    cyclesSinceLastMessage++;

  // ------- COACH : determine opponent player types -------
  if( world->my.isCoach() )
    {
      determineOtherPlayerTypes();
      return true;
    }

  // ------- NO COACH : -------

  // increase age for playerPassInfo
  for( int i = 1; i <= 11; i++ )
    if( playerPassInfo[i].age < ETERNAL )
      playerPassInfo[i].age++;

  // find out how long it will take players to reach the ball
  // (and determine the 3 "most responsible ones" and the 3 "most dangerous ones") 
  determineResponsibility();

  VALUE( "Responsible players: "
	 << firstResponsiblePlayer
	 << " (" << ourTeamReachBallTime[firstResponsiblePlayer] << " steps), "
	 << secondResponsiblePlayer
	 << " (" << ourTeamReachBallTime[secondResponsiblePlayer] << " steps), "
	 << thirdResponsiblePlayer
	 << " (" << ourTeamReachBallTime[thirdResponsiblePlayer] << " steps)" );

  VALUE( "Dangerous opponents: "
	 << firstDangerousOpponent
	 << " (" << otherTeamReachBallTime[firstDangerousOpponent] << " steps), "
	 << secondDangerousOpponent
	 << " (" << otherTeamReachBallTime[secondDangerousOpponent] << " steps), "
	 << thirdDangerousOpponent
	 << " (" << otherTeamReachBallTime[thirdDangerousOpponent] << " steps)" );

  // determine ball control
  determineBallControl();

  VALUE( "Ball control: "
	 << ( ourBallControl == BC_SAFE ? "SAFE" :
	      ourBallControl == BC_DANGER ? "DANGER" :
	      ourBallControl == BC_LOST ? "LOST" : "UNKNOWN" ) );

  // tactical action evaluation for our team
  determineOurTacticalAction();

  VALUE( "Tactical action: "
	 << ( ourAction.state == TA_CONTROL ? "CONTROL" :
	      ourAction.state == TA_PASS ? "PASS" :
	      ourAction.state == TA_GOALSHOT ? "GOALSHOT" :
	      ourAction.state == TA_REACT ? "REACT" : "UNKNOWN" )
	 << ", age: " << ourAction.age
	 << ", perf: " << ourAction.performer << ", recv: " << ourAction.receiver 
	 << ", quality: " << ourAction.quality);

  // reported action is processed, so set reportedTAction invalid
  reportedTAction.quality = TQ_UNKNOWN;
  reportedTAction.state = TA_UNKNOWN;
  reportedTAction.performer = UNKNOWN;
  reportedTAction.receiver = UNKNOWN;

  // audial information is processed, so set audioTAction invalid
  audioTAction.quality = TQ_UNKNOWN;
  audioTAction.state = TA_UNKNOWN;
  audioTAction.performer = UNKNOWN;
  audioTAction.receiver = UNKNOWN;

  return true;
}


MessageType
TacticalModule::determineMessageType( const CycleAction &action, Grade passGrade, WBall &_simulatedBall )
{
  MessageType msgType = MSGTYPE_NONE;

  // first check if player tackled
  if( action.command == TACKLE_COMMAND )
    {
      msgType = MSGTYPE_TACKLE;
      VALUE( "Communicate tackle" );
    }
  // check if a module has reported a tactical action
  if( reportedTAction.quality == TQ_REPORTED && reportedTAction.state != TA_UNKNOWN )
    {
      if( communicateSimAction )
	{
	  msgType = MSGTYPE_SIM_ACTION;
	  _simulatedBall = simulatedBall;
	  VALUE( "Communicate simaction" );
	}
      else
	{
	  msgType = MSGTYPE_ACTION;
	  VALUE( "Communicate action" );
	}
    }
  // check if passmodule thinks player is a good passpartner
  else if( passGrade.is_PERFECT() || passGrade.is_GOOD() )
    {
      msgType = MSGTYPE_POSITION;
      VALUE( "Communicate player position (as passpartner)" );
    }
  // finally check if player has seen the ball
  else if( world->ball.age == 0 && world->my.distTo( world->ball ) > VISIBLE_DISTANCE &&
	   world->my.distTo( world->ball ) < 20.0 )   // ![C]
    {
      msgType = MSGTYPE_BALL;
      VALUE( "Communicate ball" );
    }

  return msgType;
}


bool
TacticalModule::feedbackAudio( SensorData data )
{
  MessageType msgType;
  bool listenToMe = false;

  decodeAudio( data.message, data.sender_num, msgType, listenToMe, audioTAction );

  VALUE( "Got audio message from " << data.sender_num 
	 << " listenToMe : " << listenToMe 
	 << " attentionTo: " << attentionTo << " type : "
	 << ( msgType == MSGTYPE_POSITION ?  "Position" :
	      msgType == MSGTYPE_SIM_ACTION ? "sim_Action" :
	      msgType == MSGTYPE_TACKLE ? "tackle":
	      msgType == MSGTYPE_ACTION ? "action":
	      msgType == MSGTYPE_BALL ? "ball": "unknown" ) );

  // [MW] attentionTo stuff
  // listen to a player who says he controls the ball
  if( data.sender_num != attentionTo )
    {
      // attentionTo not set or message from different player
      if( listenToMe )
	{
	  attentionTo = data.sender_num;
	  cyclesSinceLastMessage = ETERNAL;
	  VALUE( "setting attentionTo to player: " << data.sender_num );
	  skills->attentionTo( attentionTo );
	  return true;
	}

      if( msgType == MSGTYPE_POSITION && audioTAction.state == TA_CONTROL &&
	  audioTAction.performer != world->my.number )
	{
	  // different player has reported his position. Listen to the performer
	  attentionTo = audioTAction.performer;
	  cyclesSinceLastMessage = ETERNAL;
	  VALUE( "heard position: setting attentionTo to player: " << attentionTo );
	  skills->attentionTo( attentionTo );
	  return true;
	}
    }
  else
    {
      cyclesSinceLastMessage = 0;
    }

  if( attentionTo != UNKNOWN && 
      ( attentionTo != data.sender_num || audioTAction.state != TA_CONTROL ) )
    {
      attentionTo = UNKNOWN;
      cyclesSinceLastMessage = ETERNAL;
      VALUE( "setting attentionTo off" );
      skills->attentionToOff();
      return true;
    }

  return true;
}


bool
TacticalModule::decodeAudio( const char * message, int sender, MessageType &msgType,
			     bool &listenToMe, TacticalAction &audioTAction )
{
  // get message type
  long c = msgDecodeChars[(int)message[0]];
  c += 74 * msgDecodeChars[(int)message[1]];
  c += 74 * 74 * msgDecodeChars[(int)message[2]];
  msgType = MessageType( c % 10 );

  // get information back into one long from 3 NodaChars
  c = msgDecodeChars[(int)message[6]];
  c += 74 * msgDecodeChars[(int)message[7]];
  c += 74 * 74 * msgDecodeChars[(int)message[8]];

  // decode tactical action

  // get state and quality
  audioTAction.state = TA_UNKNOWN;
  audioTAction.quality = TQ_UNKNOWN;

  if( c % 9 != 0 )
    {
      switch( ( c % 9 - 1 ) / 2 )
	{
	case 0  : audioTAction.state = TA_CONTROL; break;
	case 1  : audioTAction.state = TA_PASS; break;
	case 2  : audioTAction.state = TA_GOALSHOT; break;
	case 3  : audioTAction.state = TA_REACT; break;
	default : break;
	}

      if( ( c % 9 - 1 ) % 2 == 0 )
	audioTAction.quality = TQ_REPORTED;
      else
	audioTAction.quality = TQ_AUDIAL_REPORTED;
    }

  // get performer of action
  c /= 9;
  audioTAction.performer = c % 12 == 0 ? UNKNOWN : c % 12;

  // get receiver of action
  c /= 12;
  audioTAction.receiver = c % 12 == 0 ? UNKNOWN : c % 12;

  // get age of action
  c /= 12;
  audioTAction.age = c % 9 + 1;

  // get listenToMe flag (stamina is decoded in worldmodel)
  c /= 9;
  c /= 17;
  if( c == 1 )
    {
      listenToMe = true;
      cyclesSinceLastMessage = 0;
    }
  else listenToMe = false;

  if( msgType == MSGTYPE_TACKLE )
    {
      playerPassInfo[sender].grade.set_BAD();
      playerPassInfo[sender].xIndex = 0;
      playerPassInfo[sender].yIndex = 0;
    }
  else if( msgType == MSGTYPE_POSITION )
    {
      // get information from passmodule
      // CAUTION: feedbackWorld() is called hereafter and increases age!
      c = msgDecodeChars[(int)message[9]];
      playerPassInfo[sender].age = 0;

      switch( c )
	{
	case 72 :
	case 73 :   // player just tackled or is goalie or is a bad passpartner
	  playerPassInfo[sender].grade.set_BAD();
	  playerPassInfo[sender].xIndex = 0;
	  playerPassInfo[sender].yIndex = 0;
	  break;

	default :
	  switch( c % 4 )
	    {
	    case 0 : playerPassInfo[sender].grade.set_PERFECT(); break;
	    case 1 : playerPassInfo[sender].grade.set_GOOD(); break;
	    case 2 : playerPassInfo[sender].grade.set_AVERAGE(); break;
	    case 3 : playerPassInfo[sender].grade.set_POOR(); break;
	    }
	  c = c / 4;
	  if( c == 15 )
	    {
	      playerPassInfo[sender].xIndex = 3;
	      playerPassInfo[sender].yIndex = 0;
	    }
	  else
	    {
	      playerPassInfo[sender].xIndex = c / 5;
	      playerPassInfo[sender].yIndex = c % 5 - 2;
	    }
	  break;
	}
    }

  return true;
}


void
TacticalModule::encodeAudio( char* msg )
{
  bool listenToMe = ( reportedTAction.quality == TQ_REPORTED &&
		      reportedTAction.state == TA_CONTROL &&
		      ourTeamReachBallTime[world->my.number] < 3 ) ;

  // send tactical module information:
  // our tactical action, sending player's stamina
  // space needed: 3 NodaChars (with up to 74 values each)

  long c = 0;

  // send tactical action for our team

  // send state and quality (packed into 0..8)

  // if a module reported a tactical action use this one
  if( reportedTAction.state != TA_UNKNOWN )
    {
      switch( reportedTAction.state )
	{
	case TA_CONTROL  : c = 1; break;
	case TA_PASS     : c = 3; break;
	case TA_GOALSHOT : c = 5; break;
	case TA_REACT    : c = 7; break;
	default : break;
	}
      // send performer of action (0 for UNKNOWN, 1..11 else)
      if( reportedTAction.performer != UNKNOWN )
	c += 9 * reportedTAction.performer;

      // send receiver of action (0 for UNKNOWN, 1..11 else)
      if( reportedTAction.receiver != UNKNOWN )
	c += 9 * 12 * reportedTAction.receiver;

      // in this case age of action is zero
    }
  else
    {
      // In all these cases the receivers will not use the audio information
      if( ourAction.quality == TQ_AUDIAL ||
	  ourAction.quality == TQ_GUESSED ||
	  ourAction.quality == TQ_UNKNOWN ||
	  ourAction.state == TA_UNKNOWN )
	{
	  c = 0;
	}
      else
	{
	  switch( ourAction.state )
	    {
	    case TA_CONTROL  : c = 1; break;
	    case TA_PASS     : c = 3; break;
	    case TA_GOALSHOT : c = 5; break;
	    case TA_REACT    : c = 7; break;
	    default : break;
	    }
	  if( ourAction.quality == TQ_AUDIAL_REPORTED )
	    c += 1;
	}

      // send performer of action (0 for UNKNOWN, 1..11 else)
      if( ourAction.performer != UNKNOWN )
	c += 9 * ourAction.performer;

      // send receiver of action (0 for UNKNOWN, 1..11 else)
      if( ourAction.receiver != UNKNOWN )
	c += 9 * 12 * ourAction.receiver;

      // send age of action (0..8)
      c += 9 * 12 * 12 * ( max( 0, min( 8, ourAction.age ) ) );
    }

  // send player's stamina (0..16)
  c += 9 * 12 * 12 * 9 * max( 0, min( 16, int( world->my.stamina / STAMINA_MAX * 16 ) ) );

  if( listenToMe )
    c += 9 * 12 * 12 * 9 * 17;

  // send information in 3 NodaChar packets
  msg[6] = msgEncodeChars[(int)char( c % 74 )];
  msg[7] = msgEncodeChars[(int)char( c / 74 % 74 )];
  msg[8] = msgEncodeChars[(int)char( c / 74 / 74 % 74 )];
}


class AudioMsg& operator << ( AudioMsg &aumsg, int *playerType )
{
  for( int i = 1; i <= 11; i++ )
    aumsg << NodaChar( char( playerType[i] + 1 ) );

  return aumsg;
}
